#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Option class representing an enumerated list of possible values.
"""

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from . import Option
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging


# -------------------------------------------------------------------------
#
# EnumeratedListOption class
#
# -------------------------------------------------------------------------
class EnumeratedListOption(Option):
    """
    This class describes an option that provides a finite number of values.
    Each possible value is assigned a value and a description.
    """

    __signals__ = {"options-changed": None}

    def __init__(self, label, value):
        """
        :param label: A friendly label to be applied to this option.
            Example: "Paper Size"
        :type label: string
        :param value: An initial value for this option.
            Example: 5
        :type value: int
        :return: nothing
        """
        self.ini_value = value
        Option.__init__(self, label, value)
        self.__items = []
        self.__xml_items = []

    def add_item(self, value, description, xml_item=False):
        """
        Add an item to the list of possible values.

        :param value: The value that corresponds to this item.
            Example: 5
        :type value: int
        :param description: A description of this value.
            Example: "8.5 x 11"
        :type description: string
        :param xml_item: allow deferred translation of item description
        :type _xml_item: Boolean
        :return: nothing
        """
        if not xml_item:  # localized item description is being passed in
            self.__items.append((value, description))
        else:  # English item description is being passed in
            self.__xml_items.append((value, description))
            self.__items.append((value, _(description)))
        self.emit("options-changed")

    def set_items(self, items, xml_items=False):
        """
        Add a list of items to the list of possible values.

        :param items: A list of tuples containing value, description pairs.
            Example: [ (5,"8.5 x 11"), (6,"11 x 17")]
        :type items: array
        :param xml_items: allow deferred translation of item descriptions
        :type _xml_items: Boolean
        :return: nothing
        """
        if not xml_items:  # localized item descriptions are being passed in
            self.__items = items
        else:  # English item descriptions are being passed in
            self.__xml_items = items
            for value, description in items:
                self.__items.append((value, _(description)))
        self.emit("options-changed")

    def get_items(self, xml_items=False):
        """
        Get all the possible values for this option.

        :param xml_items: allow deferred translation of item descriptions
        :type _xml_items: Boolean
        :return: an array of tuples containing (value,description) pairs.
        """
        if not xml_items:  # localized item descriptions are wanted
            return self.__items
        return self.__xml_items  # English item descriptions are wanted

    def clear(self):
        """
        Clear all possible values from this option.

        :return: nothing.
        """
        self.__items = []
        self.emit("options-changed")

    def set_value(self, value):
        """
        Set the value of this option.

        :param value: A value for this option.
            Example: True
        :type value: The type will depend on the type of option.
        :return: nothing
        """
        if value in (v for v, d in self.__items):
            Option.set_value(self, value)
        elif value == self.ini_value:
            return
        else:
            logging.warning(
                _("Value '%(val)s' not found for option '%(opt)s'")
                % {"val": str(value), "opt": self.get_label()}
            )
            logging.warning(_("Valid values: ") + str(self.__items))



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    StatisticsPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Person, Family, Event, Place, Source, Citation, Repository
from gramps.gen.plug.report import Bibliography
from gramps.gen.utils.file import media_path_full
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR

LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8
_ = glocale.translation.sgettext


class StatisticsPage(BasePage):
    """
    Create one page for statistics
    """

    def __init__(self, report, the_lang, the_title, step):
        """
        @param: report        -- The instance of the main report class
                                 for this report
        @param: the_lang      -- The lang to process
        @param: the_title     -- The title page related to the language
        @param: step          -- Use to continue the progess bar
        """
        import os

        BasePage.__init__(self, report, the_lang, the_title)
        self.bibli = Bibliography()
        self.uplink = False
        self.report = report
        # set the file name and open file
        output_file, sio = self.report.create_file("statistics")
        result = self.write_header(_("Statistics"))
        addressbookpage, dummy_head, dummy_body, outerwrapper = result
        (males, females, others, unknown) = self.get_gender(
            report.database.iter_person_handles()
        )

        step()
        mobjects = report.database.get_number_of_media()
        npersons = report.database.get_number_of_people()
        nfamilies = report.database.get_number_of_families()
        nsurnames = len(set(report.database.surname_list))
        notfound = []
        total_media = 0
        mbytes = "0"
        chars = 0
        for media in report.database.iter_media():
            total_media += 1
            fullname = media_path_full(report.database, media.get_path())
            try:
                chars += os.path.getsize(fullname)
                length = len(str(chars))
                if chars <= 999999:
                    mbytes = self._("less than 1")
                else:
                    mbytes = str(chars)[: (length - 6)]
            except OSError:
                notfound.append(media.get_path())

        with Html("div", class_="content", id="EventDetail") as section:
            section += Html("h3", self._("Database overview"), inline=True)
        outerwrapper += section
        with Html("div", class_="content", id="subsection narrative") as sec11:
            sec11 += Html("h4", self._("Individuals"), inline=True)
        outerwrapper += sec11
        with Html("div", class_="content", id="subsection narrative") as sec1:
            sec1 += Html(
                "br",
                self._("Number of individuals") + self.colon + "%d" % npersons,
                inline=True,
            )
            sec1 += Html("br", self._("Males") + self.colon + "%d" % males, inline=True)
            sec1 += Html(
                "br", self._("Females") + self.colon + "%d" % females, inline=True
            )
            sec1 += Html(
                "br",
                self._("Individuals with other gender") + self.colon + "%d" % others,
                inline=True,
            )
            sec1 += Html(
                "br",
                self._("Individuals with unknown gender") + self.colon + "%d" % unknown,
                inline=True,
            )
        outerwrapper += sec1
        with Html("div", class_="content", id="subsection narrative") as sec2:
            sec2 += Html("h4", self._("Family Information"), inline=True)
            sec2 += Html(
                "br",
                self._("Number of families") + self.colon + "%d" % nfamilies,
                inline=True,
            )
            sec2 += Html(
                "br",
                self._("Unique surnames") + self.colon + "%d" % nsurnames,
                inline=True,
            )
        outerwrapper += sec2
        with Html("div", class_="content", id="subsection narrative") as sec3:
            sec3 += Html("h4", self._("Media Objects"), inline=True)
            sec3 += Html(
                "br",
                self._("Total number of media object references")
                + self.colon
                + "%d" % total_media,
                inline=True,
            )
            sec3 += Html(
                "br",
                self._("Number of unique media objects") + self.colon + "%d" % mobjects,
                inline=True,
            )
            sec3 += Html(
                "br",
                self._("Total size of media objects")
                + self.colon
                + "%8s %s" % (mbytes, self._("MB", "Megabyte")),
                inline=True,
            )
            sec3 += Html(
                "br",
                self._("Missing Media Objects") + self.colon + "%d" % len(notfound),
                inline=True,
            )
        outerwrapper += sec3
        with Html("div", class_="content", id="subsection narrative") as sec4:
            sec4 += Html("h4", self._("Miscellaneous"), inline=True)
            sec4 += Html(
                "br",
                self._("Number of events")
                + self.colon
                + "%d" % report.database.get_number_of_events(),
                inline=True,
            )
            sec4 += Html(
                "br",
                self._("Number of places")
                + self.colon
                + "%d" % report.database.get_number_of_places(),
                inline=True,
            )
            nsources = report.database.get_number_of_sources()
            sec4 += Html(
                "br",
                self._("Number of sources") + self.colon + "%d" % nsources,
                inline=True,
            )
            ncitations = report.database.get_number_of_citations()
            sec4 += Html(
                "br",
                self._("Number of citations") + self.colon + "%d" % ncitations,
                inline=True,
            )
            nrepo = report.database.get_number_of_repositories()
            sec4 += Html(
                "br",
                self._("Number of repositories") + self.colon + "%d" % nrepo,
                inline=True,
            )
        outerwrapper += sec4

        (males, females, others, unknown) = self.get_gender(
            self.report.bkref_dict[Person].keys()
        )

        origin = " :<br/>" + report.filter.get_name(self.rlocale)
        with Html("div", class_="content", id="EventDetail") as section:
            section += Html(
                "h3", self._("Narrative web content report for") + origin, inline=True
            )
        outerwrapper += section
        with Html("div", class_="content", id="subsection narrative") as sec5:
            sec5 += Html("h4", self._("Individuals"), inline=True)
            sec5 += Html(
                "br",
                self._("Number of individuals")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Person]),
                inline=True,
            )
            sec5 += Html("br", self._("Males") + self.colon + "%d" % males, inline=True)
            sec5 += Html(
                "br", self._("Females") + self.colon + "%d" % females, inline=True
            )
            sec5 += Html(
                "br",
                self._("Individuals with other gender") + self.colon + "%d" % others,
                inline=True,
            )
            sec5 += Html(
                "br",
                self._("Individuals with unknown gender") + self.colon + "%d" % unknown,
                inline=True,
            )
        outerwrapper += sec5
        with Html("div", class_="content", id="subsection narrative") as sec6:
            sec6 += Html("h4", self._("Family Information"), inline=True)
            sec6 += Html(
                "br",
                self._("Number of families")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Family]),
                inline=True,
            )
        outerwrapper += sec6
        with Html("div", class_="content", id="subsection narrative") as sec7:
            sec7 += Html("h4", self._("Miscellaneous"), inline=True)
            sec7 += Html(
                "br",
                self._("Number of events")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Event]),
                inline=True,
            )
            sec7 += Html(
                "br",
                self._("Number of places")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Place]),
                inline=True,
            )
            sec7 += Html(
                "br",
                self._("Number of sources")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Source]),
                inline=True,
            )
            sec7 += Html(
                "br",
                self._("Number of citations")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Citation]),
                inline=True,
            )
            sec7 += Html(
                "br",
                self._("Number of repositories")
                + self.colon
                + "%d" % len(self.report.bkref_dict[Repository]),
                inline=True,
            )
        outerwrapper += sec7

        # add fullclear for proper styling
        # and footer section to page
        footer = self.write_footer(None)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(addressbookpage, output_file, sio, 0)

    def get_gender(self, person_list):
        """
        This function return the number of males, females, others and unknown
        gender from a person list.

        @param: person_list -- The list to process
        """
        males = 0
        females = 0
        others = 0
        unknown = 0
        for person_handle in person_list:
            person = self.report.database.get_person_from_handle(person_handle)
            gender = person.get_gender()
            if gender == Person.MALE:
                males += 1
            elif gender == Person.FEMALE:
                females += 1
            elif gender == Person.OTHER:
                others += 1
            else:
                unknown += 1
        return (males, females, others, unknown)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SrcAttrModel
#
# -------------------------------------------------------------------------
class SrcAttrModel(Gtk.ListStore):
    def __init__(self, attr_list, db):
        Gtk.ListStore.__init__(self, str, str, bool, object)
        self.db = db
        for attr in attr_list:
            self.append(
                row=[
                    str(attr.get_type()),
                    attr.get_value(),
                    attr.get_privacy(),
                    attr,
                ]
            )



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Slovak-specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
# GRAMPS 3.x - Slovak-specific terms by Lubo Vasko
#
# -------------------------------------------------------------------------

# hĺbka použitá pre označenie / zistenie vzťahov od genenácie :
# ku generácii,


_level_name = [
    "prvého",
    "druhého",
    "tretieho",
    "štvrtého",
    "piateho",
    "šiesteho",
    "siedmeho",
    "ôsmeho",
    "deviateho",
    "desiateho",
    "jedenásteho",
    "dvanásteho",
    "trinásteho",
    "štrnásteho",
    "pätnásteho",
    "šestnásteho",
    "sedemnásteho",
    "osemnásteho",
    "devätnásteho",
    "dvadsiateho",
    "dvadsiatehoprvého",
    "dvadsiatehodruhého",
    "dvadsiatehotretieho",
    "dvadsiatehoštvrtého",
    "dvadsiatehopiateho",
    "dvadsiatehošiesteho",
    "dvadsiatehosiedmeho",
    "dvadsiatehoôsmeho",
    "dvadsiatehodeviateho",
    "tridsiateho",
]


# vzdialení príbuzní

_removed_level = [
    "prvého",
    "druhého",
    "tretieho",
    "štvrtého",
    "piateho",
    "šiesteho",
    "siedmeho",
    "ôsmeho",
    "deviateho",
    "desiateho",
    "jedenásteho",
    "dvanásteho",
    "trinásteho",
    "štrnásteho",
    "pätnásteho",
    "šestnásteho",
    "sedemnásteho",
    "osemnásteho",
    "devätnásteho",
    "dvadsiateho",
    "dvadsiatehoprvého",
    "dvadsiatehodruhého",
    "dvadsiatehotretieho",
    "dvadsiatehoštvrtého",
    "dvadsiatehopiateho",
    "dvadsiatehošiesteho",
    "dvadsiatehosiedmeho",
    "dvadsiatehoôsmeho",
    "dvadsiatehodeviateho",
    "tridsiateho",
]


# small lists, use generation level if > [5]

_father_level = [
    "",
    "otec%s",
    "starý otec%s",
    "prastarý otec%s",
    "prapredok%s",
]

_mother_level = [
    "",
    "matka%s",
    "stará matka%s",
    "prastará matka%s",
    "prapredok%s",
]

_son_level = [
    "",
    "syn%s",
    "vnuk%s",
    "pravnuk%s",
]

_daughter_level = [
    "",
    "dcéra%s",
    "vnučka%s",
    "pravnučka%s",
]

_sister_level = [
    "",
    "sestra%s",
    "teta%s",
    "prateta%s",
    "praprateta%s",
]

_brother_level = [
    "",
    "brat%s",
    "strýko%s",
    "prastrýko%s",
    "praprastrýko%s",
]

_nephew_level = [
    "",
    "synovec%s",
    "prasynovec%s",
    "praprasynovec%s",
]

_niece_level = [
    "",
    "neter%s",
    "praneter%s",
    "prapraneter%s",
]

# kinship report

_parents_level = [
    "",
    "rodičia",
    "starí rodičia",
    "prastarí rodičia",
    "predkovia",
]

_children_level = [
    "",
    "deti",
    "vnúčatá",
    "pravnúčatá",
    "pra-pravnúčatá",
]

_siblings_level = [
    "",
    "bratia a sestry",
    "strýkovia a tety",
    "prastrýkovia a pratety",
    "pra-prastrýkovia a pra-pratety",
]

_nephews_nieces_level = [
    "",
    "synovci a netere",
    "prasynovci a pranetere",
    "pra-prasynovci a pra-pranetere",
]

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    INLAW = " (m. zväzok)"

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    # od aktívnej osoby vzhľadom k spoločnému predkovi Ga=[level]
    # pre vyhodnotenie vzťahov

    def get_cousin(self, level, removed, dir="", inlaw=""):
        if removed == 0 and level < len(_level_name):
            return "bratranec %s %sstupňa" % (_removed_level[level - 1], inlaw)
        elif (level) < (removed):
            rel_str = self.get_uncle(level - 1, inlaw)
        else:
            # limitation gen = 29
            return "vzdialený bratranec, spojený s %s generáciou" % (
                _level_name[removed]
            )

    def get_cousine(self, level, removed, dir="", inlaw=""):
        if removed == 0 and level < len(_level_name):
            return "sesternica %s %sstupňa" % (_level_name[level - 1], inlaw)
        elif (level) < (removed):
            rel_str = self.get_aunt(level - 1, inlaw)
        else:
            return "vzdialená sesternica, spojená s %s generáciou" % (
                _level_name[removed]
            )

    def get_parents(self, level):
        if level > len(_parents_level) - 1:
            return "vzdialení predkovia z %s generácie" % (_level_name[level])
        else:
            return _parents_level[level]

    def get_father(self, level, inlaw=""):
        if level > len(_father_level) - 1:
            return "vzdialený predok z %s generácie" % (_level_name[level])
        else:
            return _father_level[level] % inlaw

    def get_mother(self, level, inlaw=""):
        if level > len(_mother_level) - 1:
            return "vzdialený príbuzný, predok z %s generácie" % (_level_name[level])
        else:
            return _mother_level[level] % inlaw

    def get_parent_unknown(self, level, inlaw=""):
        if level > len(_level_name) - 1:
            return "vzdialený príbuzný, predok z %s generácie" % (_level_name[level])
        else:
            return "vzdialený príbuzný%s" % (inlaw)

    def get_son(self, level, inlaw=""):
        if level > len(_son_level) - 1:
            return "vzdialený potomok z %s generácie" % (_level_name[level + 1])
        else:
            return _son_level[level] % (inlaw)

    def get_daughter(self, level, inlaw=""):
        if level > len(_daughter_level) - 1:
            return "vzdialený potomok z %s generácie" % (_level_name[level + 1])
        else:
            return _daughter_level[level] % (inlaw)

    def get_child_unknown(self, level, inlaw=""):
        if level > len(_level_name) - 1:
            return "vzdialený potomok z %s generácie" % (_level_name[level + 1])
        else:
            return "vzdialený potomok%s" % (inlaw)

    def get_sibling_unknown(self, level, inlaw=""):
        return "vzdialený príbuzný%s" % (inlaw)

    def get_uncle(self, level, inlaw=""):
        if level > len(_brother_level) - 1:
            return "vzdialený strýko z %s generácie" % (_level_name[level])
        else:
            return _brother_level[level] % (inlaw)

    def get_aunt(self, level, inlaw=""):
        if level > len(_sister_level) - 1:
            return "vzdialená teta z %s generácie" % (_level_name[level])
        else:
            return _sister_level[level] % (inlaw)

    def get_nephew(self, level, inlaw=""):
        if level > len(_nephew_level) - 1:
            return "vzdialený synovec z %s generácie" % (_level_name[level])
        else:
            return _nephew_level[level] % (inlaw)

    def get_niece(self, level, inlaw=""):
        if level > len(_niece_level) - 1:
            return "vzdialená neter z %s generácie" % (_level_name[level])
        else:
            return _niece_level[level] % (inlaw)

    # kinship report

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        see relationship.py
        """
        rel_str = "vzdialení príbuzní"
        gen = " z %s-ej generácie"
        bygen = " na %-u generáciu"
        cmt = " (bratia alebo sestry predka" + gen % (Ga) + ")"
        if Ga == 0:
            # These are descendants
            if Gb < len(_children_level):
                rel_str = _children_level[Gb]
            else:
                rel_str = "potomkovia" + gen % (Gb + 1)
        elif Gb == 0:
            # These are parents/grand parents
            if Ga < len(_parents_level):
                rel_str = _parents_level[Ga]
            else:
                rel_str = "predkovia" + gen % (Ga + 1)
        elif Gb == 1:
            # These are siblings/aunts/uncles
            if Ga < len(_siblings_level):
                rel_str = _siblings_level[Ga]
            else:
                rel_str = "deti predka" + gen % (Ga + 1) + cmt
        elif Ga == 1:
            # These are nieces/nephews
            if Gb < len(_nephews_nieces_level):
                rel_str = _nephews_nieces_level[Gb - 1]
            else:
                rel_str = "synovci a netere" + gen % (Gb)
        elif Ga > 1 and Ga == Gb:
            # These are cousins in the same generation
            # use custom level for latin words
            if Ga == 2:
                rel_str = "vlastní bratranci a sesternice"
            elif Ga <= len(_level_name):
                # %ss for plural
                rel_str = " %ss bratranci a sesternice" % _level_name[Ga - 2]
            # security
            else:
                rel_str = "bratranci a sesternice"
        elif Ga > 1 and Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            # use custom level for latin words and specific relation
            if Ga == 3 and Gb == 2:
                desc = " (vlastní bratranci niektorého z rodičov)"
                rel_str = "strýkovia a tety z ďalšieho kolena" + desc
            elif (
                Gb <= len(_level_name)
                and (Ga - Gb) < len(_removed_level)
                and (Ga + Gb + 1) < len(_removed_level)
            ):
                can = " z %s do %s stupňa (kan.)" % (
                    _removed_level[Gb],
                    _removed_level[Ga],
                )
                civ = " a do %s stupňa (civ.)" % (_removed_level[Ga + Gb + 1])
                rel_str = "strýkovia a tety" + can + civ
            elif Ga < len(_level_name):
                rel_str = "prastrýkovia a pratety" + bygen % (Ga + 1)
        elif Gb > 1 and Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            # use custom level for latin words and specific relation
            if Ga == 2 and Gb == 3:
                info = " (potomok bratranca-sesternice)"
                rel_str = "synovci a netere z ďalšieho kolena" + info
            elif (
                Ga <= len(_level_name)
                and (Gb - Ga) < len(_removed_level)
                and (Ga + Gb + 1) < len(_removed_level)
            ):
                can = " z %s do %s stupňa (kan.)" % (
                    _removed_level[Gb],
                    _removed_level[Ga],
                )
                civ = " a do %s stupňa (civ.)" % (_removed_level[Ga + Gb + 1])
                rel_str = "synovci a netere" + can + civ
            elif Ga < len(_level_name):
                rel_str = "synovci a netere" + bygen % (Gb)
        if in_law_b == True:
            # TODO: Translate this!
            rel_str = "spouses of %s" % rel_str

        return rel_str

    # quick report (missing on RelCalc tool - Status Bar)

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        see relationship.py
        """
        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        rel_str = "vzdialený príbuznýs%s" % (inlaw)
        bygen = " z %s generácie"
        if Ga == 0:
            # b is descendant of a
            if Gb == 0:
                rel_str = "tá istá osoba"
            elif gender_b == Person.MALE and Gb < len(_son_level):
                # spouse of daughter
                if inlaw and Gb == 1 and not step:
                    rel_str = "zať"
                else:
                    rel_str = self.get_son(Gb)
            elif gender_b == Person.FEMALE and Gb < len(_daughter_level):
                # spouse of son
                if inlaw and Gb == 1 and not step:
                    rel_str = "nevesta"
                else:
                    rel_str = self.get_daughter(Gb)
            # don't display inlaw
            elif Gb < len(_level_name) and gender_b == Person.MALE:
                rel_str = "vzdialený potomok (%d generácia)" % (Gb + 1)
            elif Gb < len(_level_name) and gender_b == Person.FEMALE:
                rel_str = "vzdialený potomok(žena) (%d generácia)" % (Gb + 1)
            else:
                return self.get_child_unknown(Gb)
        elif Gb == 0:
            # b is parents/grand parent of a
            if gender_b == Person.MALE and Ga < len(_father_level):
                # other spouse of father (new parent)
                if Ga == 1 and inlaw and self.STEP_SIB:
                    rel_str = "svokor"
                # father of spouse (family of spouse)
                elif Ga == 1 and inlaw:
                    rel_str = "otec partnera"
                else:
                    rel_str = self.get_father(Ga, inlaw)
            elif gender_b == Person.FEMALE and Ga < len(_mother_level):
                # other spouse of mother (new parent)
                if Ga == 1 and inlaw and self.STEP_SIB:
                    rel_str = "svokra"
                # mother of spouse (family of spouse)
                elif Ga == 1 and inlaw:
                    rel_str = "matka partnera"
                else:
                    rel_str = self.get_mother(Ga, inlaw)
            elif Ga < len(_level_name) and gender_b == Person.MALE:
                rel_str = "vzdialený predok%s (%d generácia)" % (inlaw, Ga + 1)
            elif Ga < len(_level_name) and gender_b == Person.FEMALE:
                rel_str = "vzdialený predok(žena)%s (%d generácia)" % (inlaw, Ga + 1)
            else:
                return self.get_parent_unknown(Ga, inlaw)
        elif Gb == 1:
            # b is sibling/aunt/uncle of a
            if gender_b == Person.MALE and Ga < len(_brother_level):
                rel_str = self.get_uncle(Ga, inlaw)
            elif gender_b == Person.FEMALE and Ga < len(_sister_level):
                rel_str = self.get_aunt(Ga, inlaw)
            else:
                # don't display inlaw
                if gender_b == Person.MALE:
                    rel_str = "vzdialený strýko" + bygen % (Ga + 1)
                elif gender_b == Person.FEMALE:
                    rel_str = "vzdialená teta" + bygen % (Ga + 1)
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
        elif Ga == 1:
            # b is niece/nephew of a
            if gender_b == Person.MALE and Gb < len(_nephew_level):
                rel_str = self.get_nephew(Gb - 1, inlaw)
            elif gender_b == Person.FEMALE and Gb < len(_niece_level):
                rel_str = self.get_niece(Gb - 1, inlaw)
            else:
                if gender_b == Person.MALE:
                    rel_str = "vzdialený synovec%s (%d generácia)" % (inlaw, Gb)
                elif gender_b == Person.FEMALE:
                    rel_str = "vzdialená neter%s (%d generácia)" % (inlaw, Gb)
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
        elif Ga == Gb:
            # a and b cousins in the same generation
            if gender_b == Person.MALE:
                rel_str = self.get_cousin(Ga - 1, 0, dir="", inlaw=inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_cousine(Ga - 1, 0, dir="", inlaw=inlaw)
            elif gender_b == Person.UNKNOWN:
                rel_str = self.get_sibling_unknown(Ga - 1, inlaw)
            else:
                return rel_str
        elif Ga > 1 and Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            if Ga == 3 and Gb == 2:
                if gender_b == Person.MALE:
                    desc = " (bratranec niektorého z rodičov)"
                    rel_str = "strýko z druhého kolena" + desc
                elif gender_b == Person.FEMALE:
                    desc = " (sesternica niektorého z rodičov)"
                    rel_str = "teta z druhého kolena" + desc
                elif gender_b == Person.UNKNOWN:
                    return self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
            elif (
                Gb <= len(_level_name)
                and (Ga - Gb) < len(_removed_level)
                and (Ga + Gb + 1) < len(_removed_level)
            ):
                can = " z %s do %s stupňa (kan.)" % (
                    _removed_level[Gb],
                    _removed_level[Ga],
                )
                civ = " a do %s stupňa (civ.)" % (_removed_level[Ga + Gb + 1])
                if gender_b == Person.MALE:
                    rel_str = "strýko" + can + civ
                elif gender_b == Person.FEMALE:
                    rel_str = "teta" + can + civ
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
            else:
                if gender_b == Person.MALE:
                    rel_str = self.get_uncle(Ga, inlaw)
                elif gender_b == Person.FEMALE:
                    rel_str = self.get_aunt(Ga, inlaw)
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
        elif Gb > 1 and Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            if Ga == 2 and Gb == 3:
                info = " (potomok bratranca/sesternice)"
                if gender_b == Person.MALE:
                    rel_str = "synovec z druhého kolena" + info
                elif gender_b == Person.FEMALE:
                    rel_str = "neter z druhého kolena" + info
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
            elif (
                Ga <= len(_level_name)
                and (Gb - Ga) < len(_removed_level)
                and (Ga + Gb + 1) < len(_removed_level)
            ):
                can = " z %s do %s stupňa (kan.)" % (
                    _removed_level[Gb],
                    _removed_level[Ga],
                )
                civ = " a do %s stupňa (civ.)" % (_removed_level[Ga + Gb + 1])
                if gender_b == Person.MALE:
                    rel_str = "synovec" + can + civ
                if gender_b == Person.FEMALE:
                    rel_str = "neter" + can + civ
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
            elif Ga > len(_level_name):
                return rel_str
            else:
                if gender_b == Person.MALE:
                    rel_str = self.get_nephew(Ga, inlaw)
                elif gender_b == Person.FEMALE:
                    rel_str = self.get_niece(Ga, inlaw)
                elif gender_b == Person.UNKNOWN:
                    rel_str = self.get_sibling_unknown(Ga, inlaw)
                else:
                    return rel_str
        return rel_str

    # RelCalc tool - Status Bar

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if sib_type == self.NORM_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "brat (vlastný)"
                elif gender_b == Person.FEMALE:
                    rel_str = "sestra (vlastná)"
                else:
                    rel_str = "vlastný brat alebo sestra"
            else:
                if gender_b == Person.MALE:
                    rel_str = "švagor"
                elif gender_b == Person.FEMALE:
                    rel_str = "švagriná"
                else:
                    rel_str = "švagor alebo švagriná"
        elif sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "brat"
                elif gender_b == Person.FEMALE:
                    rel_str = "sestra"
                else:
                    rel_str = "brat alebo sestra"
            else:
                if gender_b == Person.MALE:
                    rel_str = "švagor"
                elif gender_b == Person.FEMALE:
                    rel_str = "švagriná"
                else:
                    rel_str = "švagor alebo švagriná"
        # oznacenie vyberu spolocny otec, rev.
        elif sib_type == self.HALF_SIB_MOTHER:
            if gender_b == Person.MALE:
                rel_str = "nevlastný brat -spoloč.otec"
            elif gender_b == Person.FEMALE:
                rel_str = "nevlastná sestra -spoloč.otec"
            else:
                rel_str = "nevlastný brat alebo sestra -spoloč.otec"
        # oznacenie vyberu spolocna matka, rev.
        elif sib_type == self.HALF_SIB_FATHER:
            if gender_b == Person.MALE:
                rel_str = "nevlastný brat -spoloč.matka"
            elif gender_b == Person.FEMALE:
                rel_str = "nevlastná sestra -spoloč.matka"
            else:
                rel_str = "nevlastný brat alebo sestra -spoloč.matka"
        elif sib_type == self.STEP_SIB:
            if gender_b == Person.MALE:
                rel_str = "nevlastný brat"
            elif gender_b == Person.FEMALE:
                rel_str = "nevlastná sestra"
            else:
                rel_str = "nevlastný brat alebo sestra"
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_sk.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the base class for plugin registration.
It provides an object containing data about the plugin (version, filename, ...)
and a register for the data of all plugins .
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging
import os
import re
import sys
import traceback

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...version import VERSION as GRAMPSVERSION, VERSION_TUPLE
from ..const import IMAGE_DIR
from ..const import GRAMPS_LOCALE as glocale
from ..utils.requirements import Requirements

_ = glocale.translation.gettext

LOG = logging.getLogger("._manager")

# Fix DEBUG for AIO built with cx_Freeze
DEBUG = __debug__ and not hasattr(sys, "frozen")

# -------------------------------------------------------------------------
#
# PluginData
#
# -------------------------------------------------------------------------

# Development status
UNSTABLE = 0
EXPERIMENTAL = 1
BETA = 2
STABLE = 3
STATUS = [UNSTABLE, EXPERIMENTAL, BETA, STABLE]
STATUSTEXT = {
    UNSTABLE: _("Unstable"),
    EXPERIMENTAL: _("Experimental"),
    BETA: _("Beta"),
    STABLE: _("Stable"),
}

# Intended audience
EVERYONE = 0
EXPERT = 1
DEVELOPER = 2
AUDIENCE = [EVERYONE, EXPERT, DEVELOPER]
AUDIENCETEXT = {EVERYONE: _("Everyone"), EXPERT: _("Expert"), DEVELOPER: _("Developer")}

# possible plugin types
REPORT = 0
QUICKREPORT = 1  # deprecated
QUICKVIEW = 1
TOOL = 2
IMPORT = 3
EXPORT = 4
DOCGEN = 5
GENERAL = 6
MAPSERVICE = 7
VIEW = 8
RELCALC = 9
GRAMPLET = 10
SIDEBAR = 11
DATABASE = 12
RULE = 13
THUMBNAILER = 14
CITE = 15
PTYPE = [
    REPORT,
    QUICKREPORT,
    TOOL,
    IMPORT,
    EXPORT,
    DOCGEN,
    GENERAL,
    MAPSERVICE,
    VIEW,
    RELCALC,
    GRAMPLET,
    SIDEBAR,
    DATABASE,
    RULE,
    THUMBNAILER,
    CITE,
]
PTYPE_STR = {
    REPORT: _("Report"),
    QUICKREPORT: _("Quick view"),
    TOOL: _("Tool"),
    IMPORT: _("Importer"),
    EXPORT: _("Exporter"),
    DOCGEN: _("Document creator"),
    GENERAL: _("Plugin library"),
    MAPSERVICE: _("Map service"),
    VIEW: _("View"),
    RELCALC: _("Relationship calculator"),
    GRAMPLET: _("Gramplet"),
    SIDEBAR: _("Sidebar"),
    DATABASE: _("Database"),
    RULE: _("Rule"),
    THUMBNAILER: _("Thumbnailer"),
    CITE: _("Citation formatter"),
}

# possible report categories
CATEGORY_TEXT = 0
CATEGORY_DRAW = 1
CATEGORY_CODE = 2
CATEGORY_WEB = 3
CATEGORY_BOOK = 4
CATEGORY_GRAPHVIZ = 5
CATEGORY_TREE = 6
REPORT_CAT = [
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_CODE,
    CATEGORY_WEB,
    CATEGORY_BOOK,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
]
# possible tool categories
TOOL_DEBUG = -1
TOOL_ANAL = 0
TOOL_DBPROC = 1
TOOL_DBFIX = 2
TOOL_REVCTL = 3
TOOL_UTILS = 4
TOOL_CAT = [TOOL_DEBUG, TOOL_ANAL, TOOL_DBPROC, TOOL_DBFIX, TOOL_REVCTL, TOOL_UTILS]

# possible quickreport categories
CATEGORY_QR_MISC = -1
CATEGORY_QR_PERSON = 0
CATEGORY_QR_FAMILY = 1
CATEGORY_QR_EVENT = 2
CATEGORY_QR_SOURCE = 3
CATEGORY_QR_PLACE = 4
CATEGORY_QR_REPOSITORY = 5
CATEGORY_QR_NOTE = 6
CATEGORY_QR_DATE = 7
CATEGORY_QR_MEDIA = 8
CATEGORY_QR_CITATION = 9
CATEGORY_QR_SOURCE_OR_CITATION = 10

# Modes for generating reports
REPORT_MODE_GUI = 1  # Standalone report using GUI
REPORT_MODE_BKI = 2  # Book Item interface using GUI
REPORT_MODE_CLI = 4  # Command line interface (CLI)
REPORT_MODES = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]

# Modes for running tools
TOOL_MODE_GUI = 1  # Standard tool using GUI
TOOL_MODE_CLI = 2  # Command line interface (CLI)
TOOL_MODES = [TOOL_MODE_GUI, TOOL_MODE_CLI]

# possible view orders
START = 1
END = 2


# -------------------------------------------------------------------------
#
# Functions and classes
#
# -------------------------------------------------------------------------
def myint(string):
    """
    Protected version of int()
    """
    try:
        value = int(string)
    except ValueError:
        value = string
    return value


def version(string_version):
    """
    Return the tuple version of a string version.
    """
    return tuple(myint(x or "0") for x in (f"{string_version}..").split("."))


def valid_plugin_version(plugin_version_string):
    """
    Checks to see if string is a valid version string for this version
    of Gramps.
    """
    if not isinstance(plugin_version_string, str):
        return False
    dots = plugin_version_string.count(".")
    if dots == 1:
        plugin_version = tuple(map(int, plugin_version_string.split(".", 1)))
        return plugin_version == VERSION_TUPLE[:2]
    if dots == 2:
        plugin_version = tuple(map(int, plugin_version_string.split(".", 2)))
        return (
            plugin_version[:2] == VERSION_TUPLE[:2] and plugin_version <= VERSION_TUPLE
        )
    return False


class PluginData:
    """
    This is the base class for all plugin data objects.
    The workflow is:

    1. plugin manager reads all register files, and stores plugin data
       objects in a plugin register
    2. when plugin is needed, the plugin register creates the plugin, and
       the manager stores this, after which it can be executed.

    Attributes present for all plugins

    .. attribute:: id
       A unique identifier for the plugin. This is eg used to store the plugin
       settings.  MUST be in ASCII, with only "_- ().,'" special characters.
    .. attribute:: name
       A friendly name to call this plugin (normally translated)
    .. attribute:: name_accell
       A friendly name to call this plugin (normally translated), with an
       accellerator present (eg '_Descendant report', with D to be accellerator
       key
    .. attribute:: description
       A friendly description of what the plugin does
    .. attribute:: version
       The version of the plugin
    .. attribute:: status
       The development status of the plugin, UNSTABLE, EXPERIMENTAL, BETA or
       STABLE. UNSTABLE is only visible in development code, not in release
    .. attribute:: audience
       The intended audience of the plugin, EVERYONE, DEVELOPER or EXPERT.
    .. attribute:: fname
       The python file where the plugin implementation can be found
    .. attribute:: fpath
       The python path where the plugin implementation can be found
    .. attribute:: ptype
       The plugin type. One of REPORT , QUICKREPORT, TOOL, IMPORT,
        EXPORT, DOCGEN, GENERAL, MAPSERVICE, VIEW, GRAMPLET, DATABASE, RULE
    .. attribute:: authors
       List of authors of the plugin, default=[]
    .. attribute:: authors_email
       List of emails of the authors of the plugin, default=[]
    .. attribute:: maintainers
       List of maintainers of the plugin, default=[]
    .. attribute:: maintainers_email
       List of emails of the maintainers of the plugin, default=[]
    .. attribute:: supported
       Bool value indicating if the plugin is still supported, default=True
    .. attribute:: load_on_reg
       bool value, if True, the plugin is loaded on Gramps startup. Some
       plugins. Only set this value if for testing you want the plugin to be
       loaded immediately on startup. default=False
    .. attribute: icons
       New stock icons to register. A list of tuples (stock_id, icon_label),
       eg:
            [('gramps_myplugin', _('My Plugin')),
            ('gramps_myplugin_open', _('Open Plugin')]
       The icon directory must contain the directories scalable, 48x48, 22x22
       and 16x16 with the icons, eg:
            scalable/gramps_myplugin.svg
            48x48/gramps_myplugin.png
            22x22/gramps_myplugin.png
    .. attribute: icondir
       The directory to use for the icons. If icondir is not set or None, it
       reverts to the plugindirectory itself.
    .. attribute:: help_url
       The URL where documentation for the plugin can be found
    .. attribute:: requires_mod
       A list of required modules that should be importable using the python
       `import` statement.
    .. attribute:: requires_gi
       A list of (module, version) tuples that specify required modules that
       are reuired to be loaded via the GObject introspection system.
       eg: [('GExiv2', '0.10')]
    .. attribute:: requires_exe
       A list of executables required by the plugin. These are searched for in
       the paths specified in the PATH environment variable.

    Attributes for RELCALC plugins:

    .. attribute:: relcalcclass
       The class in the module that is the relationcalc class
    .. attribute:: lang_list
       List of languages this plugin handles

    Attributes for REPORT plugins:

    .. attribute:: require_active
       Bool, If the reports requries an active person to be set or not
    .. attribute:: reportclass
       The class in the module that is the report class
    .. attribute:: report_modes
       The report modes: list of REPORT_MODE_GUI ,REPORT_MODE_BKI,REPORT_MODE_CLI

    Attributes for REPORT and TOOL and QUICKREPORT and VIEW plugins

    .. attribute:: category
       Or the report category the plugin belongs to, default=CATEGORY_TEXT
       or the tool category a plugin belongs to, default=TOOL_UTILS
       or the quickreport category a plugin belongs to, default=CATEGORY_QR_PERSON
       or the view category a plugin belongs to,
           default=("Miscellaneous", _("Miscellaneous"))

    Attributes for REPORT and TOOL and DOCGEN plugins

    .. attribute:: optionclass
       The class in the module that is the option class

    Attributes for TOOL plugins

    .. attribute:: toolclass
       The class in the module that is the tool class
    .. attribute:: tool_modes
       The tool modes: list of TOOL_MODE_GUI, TOOL_MODE_CLI

    Attributes for DOCGEN plugins

    .. attribute :: docclass
       The class in the module that is the BaseDoc defined
    .. attribute :: paper
       bool, Indicates whether the plugin uses paper or not, default=True
    .. attribute :: style
       bool, Indicates whether the plugin uses styles or not, default=True

    Attribute for DOCGEN, EXPORT plugins

    .. attribute :: extension
       str, The file extension to use for output produced by the docgen/export,
       default=''

    Attributes for QUICKREPORT plugins

    .. attribute:: runfunc
       The function that executes the quick report

    Attributes for MAPSERVICE plugins

    .. attribute:: mapservice
       The class in the module that is a mapservice

    Attributes for EXPORT plugins

    .. attribute:: export_function
       Function that produces the export
    .. attribute:: export_options
       Class to set options
    .. attribute:: export_options_title
       Title for the option page

    Attributes for IMPORT plugins

    .. attribute:: import_function
       Function that starts an import

    Attributes for GRAMPLET plugins

    .. attribute:: gramplet
       The function or class that defines the gramplet.
    .. attribute:: height
       The height the gramplet should have in a column on GrampletView,
       default = 200
    .. attribute:: detached_height
       The height the gramplet should have detached, default 300
    .. attribute:: detached_width
       The width the gramplet should have detached, default 400
    .. attribute:: expand
       If the attributed should be expanded on start, default False
    .. attribute:: gramplet_title
       Title to use for the gramplet, default = _('Gramplet')
    .. attribute:: navtypes
       Navigation types that the gramplet is appropriate for, default = []

    Attributes for VIEW plugins

    .. attribute:: viewclass
       A class of type ViewCreator that holds the needed info of the
       view to be created: icon, viewclass that derives from pageview, ...
    .. attribute:: stock_category_icon
       The icon used for the view category if there is none
    .. attribute:: stock_icon
       The icon in the toolbar or sidebar used to select the view

    Attributes for SIDEBAR plugins

    .. attribute:: sidebarclass
       The class that defines the sidebar.
    .. attribute:: menu_label
       A label to use on the seltion menu.

    Attributes for VIEW and SIDEBAR plugins

    .. attribute:: order
       order can be START or END. Default is END. For END, on registering,
       the plugin is appended to the list of plugins. If START, then the
       plugin is prepended. Only set START if you want a plugin to be the
       first in the order of plugins

    Attributes for DATABASE plugins

    .. attribute:: databaseclass
       The class in the module that is the database class
    .. attribute:: reset_system
       Boolean to indicate that the system (sys.modules) should
       be reset.

    Attributes for RULE plugins

    .. attribute:: namespace
       The class (Person, Event, Media, etc.) the rule applies to.
    .. attribute:: ruleclass
       The exact class name of the rule; ex: HasSourceParameter

    Attributes for THUMBNAILER plugins

    .. attribute:: thumbnailer
       The exact class name of the thumbnailer
    """

    def __init__(self):
        # read/write attribute
        self.directory = None
        # base attributes
        self._id = None
        self._name = None
        self._name_accell = None
        self._version = None
        self._gramps_target_version = None
        self._description = None
        self._status = UNSTABLE
        self._audience = EVERYONE
        self._fname = None
        self._fpath = None
        self._ptype = None
        self._authors = []
        self._authors_email = []
        self._maintainers = []
        self._maintainers_email = []
        self._supported = True
        self._load_on_reg = False
        self._icons = []
        self._icondir = None
        self._depends_on = []
        self._requires_mod = []
        self._requires_gi = []
        self._requires_exe = []
        self._include_in_listing = True
        self._help_url = None
        # derived var
        self.mod_name = None
        # RELCALC attr
        self._relcalcclass = None
        self._lang_list = None
        # REPORT attr
        self._reportclass = None
        self._require_active = True
        self._report_modes = [REPORT_MODE_GUI]
        # REPORT and TOOL and GENERAL attr
        self._category = None
        # REPORT and TOOL attr
        self._optionclass = None
        # TOOL attr
        self._toolclass = None
        self._tool_modes = [TOOL_MODE_GUI]
        # DOCGEN attr
        self._paper = True
        self._style = True
        self._extension = ""
        # QUICKREPORT attr
        self._runfunc = None
        # MAPSERVICE attr
        self._mapservice = None
        # EXPORT attr
        self._export_function = None
        self._export_options = None
        self._export_options_title = ""
        # IMPORT attr
        self._import_function = None
        # GRAMPLET attr
        self._gramplet = None
        self._height = 200
        self._detached_height = 300
        self._detached_width = 400
        self._expand = False
        self._gramplet_title = _("Gramplet")
        self._navtypes = []
        self._orientation = None
        # VIEW attr
        self._viewclass = None
        self._stock_category_icon = None
        self._stock_icon = None
        # SIDEBAR attr
        self._sidebarclass = None
        self._menu_label = ""
        # VIEW and SIDEBAR attr
        self._order = END
        # DATABASE attr
        self._databaseclass = None
        self._reset_system = False
        # GENERAL attr
        self._data = []
        self._process = None
        # RULE attr
        self._ruleclass = None
        self._namespace = None
        # THUMBNAILER attr
        self._thumbnailer = None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, plugin_id):
        self._id = plugin_id

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def name_accell(self):
        if self._name_accell is None:
            return self._name
        return self._name_accell

    @name_accell.setter
    def name_accell(self, name):
        self._name_accell = name

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, plugin_version):
        self._version = plugin_version

    @property
    def gramps_target_version(self):
        return self._gramps_target_version

    @gramps_target_version.setter
    def gramps_target_version(self, target_version):
        self._gramps_target_version = target_version

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        if status not in STATUS:
            raise ValueError(f"plugin status cannot be {status}")
        self._status = status

    @property
    def audience(self):
        return self._audience

    @audience.setter
    def audience(self, audience):
        if audience not in AUDIENCE:
            raise ValueError(f"plugin audience cannot be {audience}")
        self._audience = audience

    @property
    def fname(self):
        return self._fname

    @fname.setter
    def fname(self, fname):
        self._fname = fname

    @property
    def fpath(self):
        return self._fpath

    @fpath.setter
    def fpath(self, fpath):
        self._fpath = fpath

    @property
    def ptype(self):
        return self._ptype

    @ptype.setter
    def ptype(self, ptype):
        if ptype not in PTYPE:
            raise ValueError(f"Plugin type cannot be {ptype}")
        if self._ptype is not None:
            raise ValueError("Plugin type may not be changed")
        self._ptype = ptype
        if self._ptype == REPORT:
            self._category = CATEGORY_TEXT
        elif self._ptype == TOOL:
            self._category = TOOL_UTILS
        elif self._ptype == QUICKREPORT:
            self._category = CATEGORY_QR_PERSON
        elif self._ptype == VIEW:
            self._category = ("Miscellaneous", _("Miscellaneous"))
        # if self._ptype == DOCGEN:
        #    self._load_on_reg = True

    @property
    def authors(self):
        return self._authors

    @authors.setter
    def authors(self, authors):
        if not authors or not isinstance(authors, list):
            return
        self._authors = authors

    @property
    def authors_email(self):
        return self._authors_email

    @authors_email.setter
    def authors_email(self, authors_email):
        if not authors_email or not isinstance(authors_email, list):
            return
        self._authors_email = authors_email

    @property
    def maintainers(self):
        return self._maintainers

    @maintainers.setter
    def maintainers(self, maintainers):
        if not maintainers or not isinstance(maintainers, list):
            return
        self._maintainers = maintainers

    @property
    def maintainers_email(self):
        return self._maintainers_email

    @maintainers_email.setter
    def maintainers_email(self, maintainers_email):
        if not maintainers_email or not isinstance(maintainers_email, list):
            return
        self._maintainers_email = maintainers_email

    @property
    def supported(self):
        return self._supported

    @supported.setter
    def supported(self, supported):
        if not isinstance(supported, bool):
            raise ValueError("Plugin must have supported=True or False")
        self._supported = supported

    @property
    def load_on_reg(self):
        return self._load_on_reg

    @load_on_reg.setter
    def load_on_reg(self, load_on_reg):
        if not isinstance(load_on_reg, bool):
            raise ValueError("Plugin must have load_on_reg=True or False")
        self._load_on_reg = load_on_reg

    @property
    def icons(self):
        return self._icons

    @icons.setter
    def icons(self, icons):
        if not isinstance(icons, list):
            raise ValueError("Plugin must have icons as a list")
        self._icons = icons

    @property
    def icondir(self):
        return self._icondir

    @icondir.setter
    def icondir(self, icondir):
        self._icondir = icondir

    @property
    def depends_on(self):
        return self._depends_on

    @depends_on.setter
    def depends_on(self, depends):
        if not isinstance(depends, list):
            raise ValueError("Plugin must have depends_on as a list")
        self._depends_on = depends

    @property
    def requires_mod(self):
        return self._requires_mod

    @requires_mod.setter
    def requires_mod(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_mod as a list")
        self._requires_mod = requires

    @property
    def requires_gi(self):
        return self._requires_gi

    @requires_gi.setter
    def requires_gi(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_gi as a list")
        self._requires_gi = requires

    @property
    def requires_exe(self):
        return self._requires_exe

    @requires_exe.setter
    def requires_exe(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_exe as a list")
        self._requires_exe = requires

    @property
    def include_in_listing(self):
        return self._include_in_listing

    @include_in_listing.setter
    def include_in_listing(self, include):
        if not isinstance(include, bool):
            raise ValueError("Plugin must have include_in_listing as a bool")
        self._include_in_listing = include

    @property
    def help_url(self):
        return self._help_url

    @help_url.setter
    def help_url(self, help_url):
        self._help_url = help_url

    def statustext(self):
        """
        Return status text.
        """
        return STATUSTEXT[self.status]

    # type specific plugin attributes

    # RELCALC attributes
    @property
    def relcalcclass(self):
        return self._relcalcclass

    @relcalcclass.setter
    def relcalcclass(self, relcalcclass):
        if self._ptype != RELCALC:
            raise ValueError("relcalcclass may only be set for RELCALC plugins")
        self._relcalcclass = relcalcclass

    @property
    def lang_list(self):
        return self._lang_list

    @lang_list.setter
    def lang_list(self, lang_list):
        if self._ptype != RELCALC:
            raise ValueError("relcalcclass may only be set for RELCALC plugins")
        self._lang_list = lang_list

    # REPORT attributes
    @property
    def require_active(self):
        return self._require_active

    @require_active.setter
    def require_active(self, require_active):
        if self._ptype != REPORT:
            raise ValueError("require_active may only be set for REPORT plugins")
        if not isinstance(require_active, bool):
            raise ValueError("Report must have require_active=True or False")
        self._require_active = require_active

    @property
    def reportclass(self):
        return self._reportclass

    @reportclass.setter
    def reportclass(self, reportclass):
        if self._ptype != REPORT:
            raise ValueError("reportclass may only be set for REPORT plugins")
        self._reportclass = reportclass

    @property
    def report_modes(self):
        return self._report_modes

    @report_modes.setter
    def report_modes(self, report_modes):
        if self._ptype != REPORT:
            raise ValueError("report_modes may only be set for REPORT plugins")
        if not isinstance(report_modes, list):
            raise ValueError("report_modes must be a list")
        self._report_modes = [x for x in report_modes if x in REPORT_MODES]
        if not self._report_modes:
            raise ValueError("report_modes not a valid list of modes")

    # REPORT or TOOL or QUICKREPORT or GENERAL attributes
    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        if self._ptype not in [REPORT, TOOL, QUICKREPORT, VIEW, GENERAL]:
            raise ValueError(
                "category may only be set for "
                "REPORT/TOOL/QUICKREPORT/VIEW/GENERAL plugins"
            )
        self._category = category

    # REPORT OR TOOL attributes
    @property
    def optionclass(self):
        return self._optionclass

    @optionclass.setter
    def optionclass(self, optionclass):
        if self._ptype not in [REPORT, TOOL, DOCGEN]:
            raise ValueError(
                "optionclass may only be set for REPORT/TOOL/DOCGEN plugins"
            )
        self._optionclass = optionclass

    # TOOL attributes
    @property
    def toolclass(self):
        return self._toolclass

    @toolclass.setter
    def toolclass(self, toolclass):
        if self._ptype != TOOL:
            raise ValueError("toolclass may only be set for TOOL plugins")
        self._toolclass = toolclass

    @property
    def tool_modes(self):
        return self._tool_modes

    @tool_modes.setter
    def tool_modes(self, tool_modes):
        if self._ptype != TOOL:
            raise ValueError("tool_modes may only be set for TOOL plugins")
        if not isinstance(tool_modes, list):
            raise ValueError("tool_modes must be a list")
        self._tool_modes = [x for x in tool_modes if x in TOOL_MODES]
        if not self._tool_modes:
            raise ValueError("tool_modes not a valid list of modes")

    # DOCGEN attributes
    @property
    def paper(self):
        return self._paper

    @paper.setter
    def paper(self, paper):
        if self._ptype != DOCGEN:
            raise ValueError("paper may only be set for DOCGEN plugins")
        if not isinstance(paper, bool):
            raise ValueError("Plugin must have paper=True or False")
        self._paper = paper

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style):
        if self._ptype != DOCGEN:
            raise ValueError("style may only be set for DOCGEN plugins")
        if not isinstance(style, bool):
            raise ValueError("Plugin must have style=True or False")
        self._style = style

    @property
    def extension(self):
        return self._extension

    @extension.setter
    def extension(self, extension):
        if self._ptype not in [DOCGEN, EXPORT, IMPORT]:
            raise ValueError(
                "extension may only be set for DOCGEN/EXPORT/IMPORT plugins"
            )
        self._extension = extension

    # QUICKREPORT attributes
    @property
    def runfunc(self):
        return self._runfunc

    @runfunc.setter
    def runfunc(self, runfunc):
        if self._ptype != QUICKREPORT:
            raise ValueError("runfunc may only be set for QUICKREPORT plugins")
        self._runfunc = runfunc

    # MAPSERVICE attributes
    @property
    def mapservice(self):
        return self._mapservice

    @mapservice.setter
    def mapservice(self, mapservice):
        if self._ptype != MAPSERVICE:
            raise ValueError("mapservice may only be set for MAPSERVICE plugins")
        self._mapservice = mapservice

    # EXPORT attributes
    @property
    def export_function(self):
        return self._export_function

    @export_function.setter
    def export_function(self, export_function):
        if self._ptype != EXPORT:
            raise ValueError("export_function may only be set for EXPORT plugins")
        self._export_function = export_function

    @property
    def export_options(self):
        return self._export_options

    @export_options.setter
    def export_options(self, export_options):
        if self._ptype != EXPORT:
            raise ValueError("export_options may only be set for EXPORT plugins")
        self._export_options = export_options

    @property
    def export_options_title(self):
        return self._export_options_title

    @export_options_title.setter
    def export_options_title(self, export_options_title):
        if self._ptype != EXPORT:
            raise ValueError("export_options_title may only be set for EXPORT plugins")
        self._export_options_title = export_options_title

    # IMPORT attributes
    @property
    def import_function(self):
        return self._import_function

    @import_function.setter
    def import_function(self, import_function):
        if self._ptype != IMPORT:
            raise ValueError("import_function may only be set for IMPORT plugins")
        self._import_function = import_function

    # GRAMPLET attributes
    @property
    def gramplet(self):
        return self._gramplet

    @gramplet.setter
    def gramplet(self, gramplet):
        if self._ptype != GRAMPLET:
            raise ValueError("gramplet may only be set for GRAMPLET plugins")
        self._gramplet = gramplet

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        if self._ptype != GRAMPLET:
            raise ValueError("height may only be set for GRAMPLET plugins")
        if not isinstance(height, int):
            raise ValueError("Plugin must have height an integer")
        self._height = height

    @property
    def detached_height(self):
        return self._detached_height

    @detached_height.setter
    def detached_height(self, detached_height):
        if self._ptype != GRAMPLET:
            raise ValueError("detached_height may only be set for GRAMPLET plugins")
        if not isinstance(detached_height, int):
            raise ValueError("Plugin must have detached_height an integer")
        self._detached_height = detached_height

    @property
    def detached_width(self):
        return self._detached_width

    @detached_width.setter
    def detached_width(self, detached_width):
        if self._ptype != GRAMPLET:
            raise ValueError("detached_width may only be set for GRAMPLET plugins")
        if not isinstance(detached_width, int):
            raise ValueError("Plugin must have detached_width an integer")
        self._detached_width = detached_width

    @property
    def expand(self):
        return self._expand

    @expand.setter
    def expand(self, expand):
        if self._ptype != GRAMPLET:
            raise ValueError("expand may only be set for GRAMPLET plugins")
        if not isinstance(expand, bool):
            raise ValueError("Plugin must have expand as a bool")
        self._expand = expand

    @property
    def gramplet_title(self):
        return self._gramplet_title

    @gramplet_title.setter
    def gramplet_title(self, gramplet_title):
        if self._ptype != GRAMPLET:
            raise ValueError("gramplet_title may only be set for GRAMPLET plugins")
        if not isinstance(gramplet_title, str):
            raise ValueError(
                f"gramplet_title is type {type(gramplet_title)}, "
                "string or unicode required"
            )
        self._gramplet_title = gramplet_title

    @property
    def navtypes(self):
        return self._navtypes

    @navtypes.setter
    def navtypes(self, navtypes):
        if self._ptype != GRAMPLET:
            raise ValueError("navtypes may only be set for GRAMPLET plugins")
        self._navtypes = navtypes

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        if self._ptype != GRAMPLET:
            raise ValueError("orientation may only be set for GRAMPLET plugins")
        self._orientation = orientation

    @property
    def viewclass(self):
        return self._viewclass

    @viewclass.setter
    def viewclass(self, viewclass):
        if self._ptype != VIEW:
            raise ValueError("viewclass may only be set for VIEW plugins")
        self._viewclass = viewclass

    @property
    def stock_category_icon(self):
        return self._stock_category_icon

    @stock_category_icon.setter
    def stock_category_icon(self, stock_category_icon):
        if self._ptype != VIEW:
            raise ValueError("stock_category_icon may only be set for VIEW plugins")
        self._stock_category_icon = stock_category_icon

    @property
    def stock_icon(self):
        return self._stock_icon

    @stock_icon.setter
    def stock_icon(self, stock_icon):
        if self._ptype != VIEW:
            raise ValueError("stock_icon may only be set for VIEW plugins")
        self._stock_icon = stock_icon

    # SIDEBAR attributes
    @property
    def sidebarclass(self):
        return self._sidebarclass

    @sidebarclass.setter
    def sidebarclass(self, sidebarclass):
        if self._ptype != SIDEBAR:
            raise ValueError("sidebarclass may only be set for SIDEBAR plugins")
        self._sidebarclass = sidebarclass

    @property
    def menu_label(self):
        return self._menu_label

    @menu_label.setter
    def menu_label(self, menu_label):
        if self._ptype != SIDEBAR:
            raise ValueError("menu_label may only be set for SIDEBAR plugins")
        self._menu_label = menu_label

    # VIEW, SIDEBAR and THUMBNAILER attributes
    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, order):
        if self._ptype not in [VIEW, SIDEBAR, THUMBNAILER]:
            raise ValueError(
                "order may only be set for VIEW/SIDEBAR/THUMBNAILER plugins"
            )
        self._order = order

    # DATABASE attributes
    @property
    def databaseclass(self):
        return self._databaseclass

    @databaseclass.setter
    def databaseclass(self, databaseclass):
        if self._ptype != DATABASE:
            raise ValueError("databaseclass may only be set for DATABASE plugins")
        self._databaseclass = databaseclass

    @property
    def reset_system(self):
        return self._reset_system

    @reset_system.setter
    def reset_system(self, reset_system):
        if self._ptype != DATABASE:
            raise ValueError("reset_system may only be set for DATABASE plugins")
        self._reset_system = reset_system

    # GENERAL attr
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if self._ptype != GENERAL:
            raise ValueError("data may only be set for GENERAL plugins")
        self._data = data

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, process):
        if self._ptype != GENERAL:
            raise ValueError("process may only be set for GENERAL plugins")
        self._process = process

    # RULE attr
    @property
    def ruleclass(self):
        return self._ruleclass

    @ruleclass.setter
    def ruleclass(self, data):
        if self._ptype != RULE:
            raise ValueError("ruleclass may only be set for RULE plugins")
        self._ruleclass = data

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, data):
        if self._ptype != RULE:
            raise ValueError("namespace may only be set for RULE plugins")
        self._namespace = data

    # THUMBNAILER attr
    @property
    def thumbnailer(self):
        return self._thumbnailer

    @thumbnailer.setter
    def thumbnailer(self, data):
        if self._ptype != THUMBNAILER:
            raise ValueError("thumbnailer may only be set for THUMBNAILER plugins")
        self._thumbnailer = data


def newplugin():
    """
    Function to create a new plugindata object, add it to list of
    registered plugins

    :returns: a newly created PluginData which is already part of the register
    """
    gpr = PluginRegister.get_instance()
    pgd = PluginData()
    gpr.add_plugindata(pgd)
    return pgd


def register(ptype, **kwargs):
    """
    Convenience function to register a new plugin using a dictionary as input.
    The register functions will call newplugin() function, and use the
    dictionary kwargs to assign data to the PluginData newplugin() created,
    as in: plugindata.key = data

    :param ptype: the plugin type, one of REPORT, TOOL, ...
    :param kwargs: dictionary with keys attributes of the plugin, and data
                   the value
    :returns: a newly created PluginData which is already part of the register
              and which has kwargs assigned as attributes
    """
    plg = newplugin()
    plg.ptype = ptype
    for prop, value in kwargs.items():
        # check it is a valid attribute with getattr
        getattr(plg, prop)
        # set the value
        setattr(plg, prop, value)
    return plg


def make_environment(**kwargs):
    """
    Prepare environment.
    """
    env = {
        "newplugin": newplugin,
        "register": register,
        "UNSTABLE": UNSTABLE,
        "EXPERIMENTAL": EXPERIMENTAL,
        "BETA": BETA,
        "STABLE": STABLE,
        "EVERYONE": EVERYONE,
        "DEVELOPER": DEVELOPER,
        "EXPERT": EXPERT,
        "REPORT": REPORT,
        "QUICKREPORT": QUICKREPORT,
        "TOOL": TOOL,
        "IMPORT": IMPORT,
        "EXPORT": EXPORT,
        "DOCGEN": DOCGEN,
        "GENERAL": GENERAL,
        "RULE": RULE,
        "MAPSERVICE": MAPSERVICE,
        "VIEW": VIEW,
        "RELCALC": RELCALC,
        "GRAMPLET": GRAMPLET,
        "SIDEBAR": SIDEBAR,
        "THUMBNAILER": THUMBNAILER,
        "CITE": CITE,
        "CATEGORY_TEXT": CATEGORY_TEXT,
        "CATEGORY_DRAW": CATEGORY_DRAW,
        "CATEGORY_CODE": CATEGORY_CODE,
        "CATEGORY_WEB": CATEGORY_WEB,
        "CATEGORY_BOOK": CATEGORY_BOOK,
        "CATEGORY_GRAPHVIZ": CATEGORY_GRAPHVIZ,
        "CATEGORY_TREE": CATEGORY_TREE,
        "TOOL_DEBUG": TOOL_DEBUG,
        "TOOL_ANAL": TOOL_ANAL,
        "TOOL_DBPROC": TOOL_DBPROC,
        "TOOL_DBFIX": TOOL_DBFIX,
        "TOOL_REVCTL": TOOL_REVCTL,
        "TOOL_UTILS": TOOL_UTILS,
        "CATEGORY_QR_MISC": CATEGORY_QR_MISC,
        "CATEGORY_QR_PERSON": CATEGORY_QR_PERSON,
        "CATEGORY_QR_FAMILY": CATEGORY_QR_FAMILY,
        "CATEGORY_QR_EVENT": CATEGORY_QR_EVENT,
        "CATEGORY_QR_SOURCE": CATEGORY_QR_SOURCE,
        "CATEGORY_QR_CITATION": CATEGORY_QR_CITATION,
        "CATEGORY_QR_SOURCE_OR_CITATION": CATEGORY_QR_SOURCE_OR_CITATION,
        "CATEGORY_QR_PLACE": CATEGORY_QR_PLACE,
        "CATEGORY_QR_MEDIA": CATEGORY_QR_MEDIA,
        "CATEGORY_QR_REPOSITORY": CATEGORY_QR_REPOSITORY,
        "CATEGORY_QR_NOTE": CATEGORY_QR_NOTE,
        "CATEGORY_QR_DATE": CATEGORY_QR_DATE,
        "REPORT_MODE_GUI": REPORT_MODE_GUI,
        "REPORT_MODE_BKI": REPORT_MODE_BKI,
        "REPORT_MODE_CLI": REPORT_MODE_CLI,
        "TOOL_MODE_GUI": TOOL_MODE_GUI,
        "TOOL_MODE_CLI": TOOL_MODE_CLI,
        "DATABASE": DATABASE,
        "GRAMPSVERSION": GRAMPSVERSION,
        "START": START,
        "END": END,
        "IMAGE_DIR": IMAGE_DIR,
    }
    env.update(kwargs)
    return env


# -------------------------------------------------------------------------
#
# PluginRegister
#
# -------------------------------------------------------------------------
class PluginRegister:
    """
    PluginRegister is a Singleton which holds plugin data

    .. attribute : stable_only
        Bool, include stable plugins only or not. Default True
    """

    __instance = None

    @staticmethod
    def get_instance():
        """Use this function to get the instance of the PluginRegister"""
        if PluginRegister.__instance is None:
            PluginRegister.__instance = 1  # Set to 1 for __init__()
            PluginRegister.__instance = PluginRegister()
        return PluginRegister.__instance

    def __init__(self):
        """This function should only be run once by get_instance()"""
        if PluginRegister.__instance != 1:
            raise Exception("This class is a singleton. Use the get_instance() method")
        self.stable_only = True
        if DEBUG:
            self.stable_only = False
        self.__plugindata = []
        self.__id_to_pdata = {}
        self.__req = Requirements()

    def add_plugindata(self, plugindata):
        """This is used to add an entry to the registration list.  The way it
        is used, this entry is not yet filled in, so we cannot use the id to
        add to the __id_to_pdata dict at this time."""
        self.__plugindata.append(plugindata)

    def scan_dir(self, directory, filenames, uistate=None):
        """
        The dir name will be scanned for plugin registration code, which will
        be loaded in :class:`PluginData` objects if they satisfy some checks.

        :returns: A list with :class:`PluginData` objects
        """
        # if the directory does not exist, do nothing
        if not (os.path.isdir(directory) or os.path.islink(directory)):
            return []

        ext = r".gpr.py"
        extlen = -len(ext)
        pymod = re.compile(r"^(.*)\.py$")

        for filename in filenames:
            if not filename[extlen:] == ext:
                continue
            lenpd = len(self.__plugindata)
            full_filename = os.path.join(directory, filename)
            try:
                with open(full_filename, "r", encoding="utf-8") as file_descriptor:
                    stream = file_descriptor.read()
            except Exception as msg:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print(msg)
                continue
            if os.path.exists(os.path.join(os.path.dirname(full_filename), "locale")):
                try:
                    local_gettext = glocale.get_addon_translator(full_filename).gettext
                except ValueError:
                    print(
                        _(
                            "WARNING: Plugin %(plugin_name)s has no translation"
                            " for any of your configured languages, using US"
                            " English instead"
                        )
                        % {"plugin_name": filename.split(".")[0]}
                    )
                    local_gettext = glocale.translation.gettext
            else:
                local_gettext = glocale.translation.gettext
            try:
                exec(
                    compile(stream, filename, "exec"),
                    make_environment(_=local_gettext),
                    {"uistate": uistate},
                )
                for pdata in self.__plugindata[lenpd:]:
                    if pdata.id in self.__id_to_pdata:
                        # reloading
                        old = self.__id_to_pdata[pdata.id]
                        self.__plugindata.remove(old)
                        lenpd -= 1
                    self.__id_to_pdata[pdata.id] = pdata
            except ValueError as msg:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print(msg)
                self.__plugindata = self.__plugindata[:lenpd]
            except:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print("".join(traceback.format_exception(*sys.exc_info())))
                self.__plugindata = self.__plugindata[:lenpd]
            # check if:
            #  1. plugin exists, if not remove, otherwise set module name
            #  2. plugin not stable, if stable_only=True, remove
            #  3. TOOL_DEBUG only if DEBUG True
            rmlist = []
            ind = lenpd - 1
            for plugin in self.__plugindata[lenpd:]:
                # LOG.warning("\nPlugin scanned %s at registration", plugin.id)
                ind += 1
                plugin.directory = directory
                if not valid_plugin_version(plugin.gramps_target_version):
                    print(
                        _(
                            "ERROR: Plugin file %(filename)s has a version of "
                            '"%(gramps_target_version)s" which is invalid for Gramps '
                            '"%(gramps_version)s".'
                            % {
                                "filename": os.path.join(directory, plugin.fname),
                                "gramps_version": GRAMPSVERSION,
                                "gramps_target_version": plugin.gramps_target_version,
                            }
                        )
                    )
                    rmlist.append(ind)
                    continue
                if not self.__req.check_plugin(plugin):
                    rmlist.append(ind)
                    continue
                if plugin.status == UNSTABLE and self.stable_only:
                    rmlist.append(ind)
                    continue
                if plugin.ptype == TOOL and plugin.category == TOOL_DEBUG and not DEBUG:
                    rmlist.append(ind)
                    continue
                if plugin.fname is None:
                    continue
                match = pymod.match(plugin.fname)
                if not match:
                    rmlist.append(ind)
                    print(
                        _(
                            "ERROR: Wrong python file %(filename)s in register file "
                            "%(regfile)s"
                        )
                        % {
                            "filename": os.path.join(directory, plugin.fname),
                            "regfile": os.path.join(directory, filename),
                        }
                    )
                    continue
                if not os.path.isfile(os.path.join(directory, plugin.fname)):
                    rmlist.append(ind)
                    print(
                        _(
                            "ERROR: Python file %(filename)s in register file "
                            "%(regfile)s does not exist"
                        )
                        % {
                            "filename": os.path.join(directory, plugin.fname),
                            "regfile": os.path.join(directory, filename),
                        }
                    )
                    continue
                module = match.groups()[0]
                plugin.mod_name = module
                plugin.fpath = directory
                # LOG.warning("\nPlugin added %s at registration", plugin.id)
            rmlist.reverse()
            for ind in rmlist:
                del self.__id_to_pdata[self.__plugindata[ind].id]
                del self.__plugindata[ind]

    def get_plugin(self, plugin_id):
        """
        Return the :class:`PluginData` for the plugin with id
        """
        assert len(self.__id_to_pdata) == len(self.__plugindata)
        # if len(self.__id_to_pdata) != len(self.__plugindata):
        #     print(len(self.__id_to_pdata), len(self.__plugindata))
        return self.__id_to_pdata.get(plugin_id, None)

    def type_plugins(self, ptype):
        """
        Return a list of :class:`PluginData` that are of type ptype
        """
        return [x for x in self.__plugindata if x.ptype == ptype]

    def report_plugins(self, gui=True):
        """
        Return a list of gui or cli :class:`PluginData` that are of type REPORT

        :param gui: bool, if True then gui plugin, otherwise cli plugin
        """
        if gui:
            return [
                x
                for x in self.type_plugins(REPORT)
                if REPORT_MODE_GUI in x.report_modes
            ]
        return [
            x for x in self.type_plugins(REPORT) if REPORT_MODE_CLI in x.report_modes
        ]

    def tool_plugins(self, gui=True):
        """
        Return a list of :class:`PluginData` that are of type TOOL
        """
        if gui:
            return [x for x in self.type_plugins(TOOL) if TOOL_MODE_GUI in x.tool_modes]
        return [x for x in self.type_plugins(TOOL) if TOOL_MODE_CLI in x.tool_modes]

    def bookitem_plugins(self):
        """
        Return a list of REPORT :class:`PluginData` that are can be used as
        bookitem
        """
        return [
            x for x in self.type_plugins(REPORT) if REPORT_MODE_BKI in x.report_modes
        ]

    def quickreport_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type QUICKREPORT
        """
        return self.type_plugins(QUICKREPORT)

    def import_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type IMPORT
        """
        return self.type_plugins(IMPORT)

    def export_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type EXPORT
        """
        return self.type_plugins(EXPORT)

    def docgen_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type DOCGEN
        """
        return self.type_plugins(DOCGEN)

    def general_plugins(self, category=None):
        """
        Return a list of :class:`PluginData` that are of type GENERAL
        """
        plugins = self.type_plugins(GENERAL)
        if category:
            return [plugin for plugin in plugins if plugin.category == category]
        return plugins

    def mapservice_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type MAPSERVICE
        """
        return self.type_plugins(MAPSERVICE)

    def view_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type VIEW
        """
        return self.type_plugins(VIEW)

    def relcalc_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type RELCALC
        """
        return self.type_plugins(RELCALC)

    def gramplet_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type GRAMPLET
        """
        return self.type_plugins(GRAMPLET)

    def sidebar_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type SIDEBAR
        """
        return self.type_plugins(SIDEBAR)

    def database_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type DATABASE
        """
        return self.type_plugins(DATABASE)

    def rule_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type RULE
        """
        return self.type_plugins(RULE)

    def thumbnailer_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type THUMBNAILER
        """
        return self.type_plugins(THUMBNAILER)

    def cite_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type CITE
        """
        return self.type_plugins(CITE)

    def filter_load_on_reg(self):
        """
        Return a list of :class:`PluginData` that have load_on_reg == True
        """
        return [x for x in self.__plugindata if x.load_on_reg is True]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Brian G. Matherly
# Copyright (C) 2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the management of databases from CLI. This includes opening, renaming,
creating, and deleting of databases.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import re
import os
import sys
import time
from urllib.parse import urlparse
from urllib.request import urlopen, url2pathname
import tempfile
import logging

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import BasePluginManager
from gramps.gen.config import config
from gramps.gen.constfunc import win
from gramps.gen.db.dbconst import DBLOGNAME, DBBACKEND
from gramps.gen.db.utils import make_database, get_dbid_from_path
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".clidbman")
_LOG = logging.getLogger(DBLOGNAME)

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
DEFAULT_TITLE = _("Family Tree")
NAME_FILE = "name.txt"
META_NAME = "meta_data.db"
UNAVAILABLE = _("Unavailable")


# -------------------------------------------------------------------------
#
# functions
#
# -------------------------------------------------------------------------
def _errordialog(title, errormessage):
    """
    Show the error. A title for the error and an errormessage
    """
    print(
        _("ERROR: %(title)s \n       %(message)s")
        % {"title": title, "message": errormessage}
    )
    sys.exit()


# -------------------------------------------------------------------------
#
# CLIDbManager
#
# -------------------------------------------------------------------------
class CLIDbManager:
    """
    Database manager without GTK functionality, allows users to create and
    open databases
    """

    IND_NAME = 0
    IND_PATH = 1
    IND_PATH_NAMEFILE = 2
    IND_TVAL_STR = 3
    IND_TVAL = 4
    IND_USE_ICON_BOOL = 5
    IND_STOCK_ID = 6

    ICON_NONE = 0
    ICON_RECOVERY = 1
    ICON_LOCK = 2
    ICON_OPEN = 3

    ICON_MAP: dict[int, str | None] = {
        ICON_NONE: None,
        ICON_RECOVERY: None,
        ICON_LOCK: None,
        ICON_OPEN: None,
    }

    ERROR = _errordialog

    def __init__(self, dbstate):
        self.dbstate = dbstate
        self.msg = None

        if dbstate and dbstate.is_open():
            self.active = dbstate.db.get_save_path()
        else:
            self.active = None

        self.current_names = []
        if dbstate:
            self._populate_cli()

    def empty(self, val):
        """
        Callback that does nothing
        """
        pass

    def get_dbdir_summary(self, dirpath, name):
        """
        dirpath: full path to database
        name: proper name of family tree

        Returns dictionary of summary item.
        Should include at least, if possible:

        _("Path")
        _("Family Tree")
        _("Last accessed")
        _("Database")
        _("Locked?")

        and these details:

        _("Number of people")
        _("Version")
        _("Schema version")
        """
        dbid = get_dbid_from_path(dirpath)
        if not self.is_locked(dirpath):
            try:
                database = make_database(dbid)
                database.load(dirpath, None, update=False)
                retval = database.get_summary()
                database.close(update=False)
            except Exception as msg:
                retval = {_("Unavailable"): str(msg)[:74] + "..."}
        else:
            retval = {_("Unavailable"): "locked"}
        retval.update(
            {
                _("Family Tree"): name,
                _("Path"): dirpath,
                _("Database"): self.get_backend_name_from_dbid(dbid),
                _("Last accessed"): time_val(dirpath)[1],
                _("Locked?"): self.is_locked(dirpath),
            }
        )
        return retval

    def get_backend_name_from_dbid(self, dbid):
        pmgr = BasePluginManager.get_instance()
        for plugin in pmgr.get_reg_databases():
            if plugin.id == dbid:
                return plugin._name
        return UNAVAILABLE

    def print_family_tree_summaries(self, database_names=None):
        """
        Prints a detailed list of the known family trees.
        """
        print(_("Gramps Family Trees:"))
        for item in self.current_names:
            (
                name,
                dirpath,
                path_name,
                last,
                tval,
                enable,
                stock_id,
                backend_type,
            ) = item
            if database_names is None or any(
                [
                    (re.match("^" + dbname + "$", name) or dbname == name)
                    for dbname in database_names
                ]
            ):
                summary = self.get_dbdir_summary(dirpath, name)
                print(_('Family Tree "%s":') % summary[_("Family Tree")])
                for item in sorted(summary):
                    if item != "Family Tree":
                        # Translators: needed for French, ignore otherwise
                        print(
                            "   "
                            + _("%(str1)s: %(str2)s")
                            % {"str1": item, "str2": summary[item]}
                        )

    def family_tree_summary(self, database_names=None):
        """
        Return a list of dictionaries of the known family trees.
        """
        # make the default directory if it does not exist
        summary_list = []
        for item in self.current_names:
            (
                name,
                dirpath,
                path_name,
                last,
                tval,
                enable,
                stock_id,
                backend_type,
            ) = item
            if database_names is None or any(
                [
                    (re.match("^" + dbname + "$", name) or dbname == name)
                    for dbname in database_names
                ]
            ):
                retval = self.get_dbdir_summary(dirpath, name)
                summary_list.append(retval)
        return summary_list

    def _populate_cli(self):
        """
        Get the list of current names in the database dir
        """
        # make the default directory if it does not exist
        dbdir = os.path.expanduser(config.get("database.path"))
        db_ok = make_dbdir(dbdir)

        self.current_names = []
        if db_ok:
            for dpath in os.listdir(dbdir):
                dirpath = os.path.join(dbdir, dpath)
                path_name = os.path.join(dirpath, NAME_FILE)
                backend_type = get_dbid_from_path(dirpath)
                if os.path.isfile(path_name):
                    with open(path_name, "r", encoding="utf8") as file:
                        name = file.readline().strip()

                    (tval, last) = time_val(dirpath)
                    (enable, stock_id) = self.icon_values(
                        dirpath, self.active, self.dbstate.is_open()
                    )

                    if stock_id == "gramps-lock":
                        last = find_locker_name(dirpath)

                    self.current_names.append(
                        (
                            name,
                            os.path.join(dbdir, dpath),
                            path_name,
                            last,
                            tval,
                            enable,
                            stock_id,
                            backend_type,
                        )
                    )

        self.current_names.sort()

    def get_family_tree_path(self, name):
        """
        Given a name, return None if name not existing or the path to the
        database if it is a known database name.
        """
        for data in self.current_names:
            if data[0] == name:
                return data[1]
        return None

    def family_tree_list(self):
        """
        Return a list of name, dirname of the known family trees
        """
        lst = [(x[0], x[1]) for x in self.current_names]
        return lst

    def __start_cursor(self, msg):
        """
        Do needed things to start import visually, eg busy cursor
        """
        print(_("Starting Import, %s") % msg)

    def __end_cursor(self):
        """
        Set end of a busy cursor
        """
        print(_("Import finished..."))

    def create_new_db_cli(self, title=None, create_db=True, dbid=None):
        """
        Create a new database.
        """
        new_path = find_next_db_dir()

        os.mkdir(new_path)
        path_name = os.path.join(new_path, NAME_FILE)

        if title is None:
            name_list = [name[0] for name in self.current_names]
            title = find_next_db_name(name_list)

        with open(path_name, "w", encoding="utf8") as name_file:
            name_file.write(title)

        if create_db:
            if dbid is None:
                dbid = config.get("database.backend")
            newdb = make_database(dbid)

        backend_path = os.path.join(new_path, DBBACKEND)
        with open(backend_path, "w", encoding="utf8") as backend_file:
            backend_file.write(dbid)

        (tval, last) = time_val(new_path)

        self.current_names.append(
            (title, new_path, path_name, last, tval, False, "", dbid)
        )
        return new_path, title

    def _create_new_db(self, title=None, dbid=None, edit_entry=False):
        """
        Create a new database, do extra stuff needed
        """
        return self.create_new_db_cli(title, dbid=dbid)

    def import_new_db(self, filename, user):
        """
        Attempt to import the provided file into a new database.
        A new database will only be created if an appropriate importer was
        found.

        :param filename: a fully-qualified path, filename, and
                         extension to open.

        :param user: a :class:`.cli.user.User` or :class:`.gui.user.User`
                     instance for managing user interaction.

        :returns: A tuple of (new_path, name) for the new database
                  or (None, None) if no import was performed.
        """
        pmgr = BasePluginManager.get_instance()
        # check to see if it isn't a filename directly:
        if not os.path.isfile(filename):
            # Allow URL names here; make temp file if necessary
            url = urlparse(filename)
            if url.scheme != "":
                if url.scheme == "file":
                    filename = url2pathname(filename[7:])
                else:
                    url_fp = urlopen(filename)  # open URL
                    # make a temp local file:
                    ext = os.path.splitext(url.path)[1]
                    fd, filename = tempfile.mkstemp(suffix=ext)
                    temp_fp = os.fdopen(fd, "w")
                    # read from URL:
                    data = url_fp.read()
                    # write locally:
                    temp_fp.write(data)
                    url_fp.close()
                    temp_fp.close()

        (name, ext) = os.path.splitext(os.path.basename(filename))
        format = ext[1:].lower()

        for plugin in pmgr.get_import_plugins():
            if format == plugin.get_extension():
                dbid = config.get("database.backend")
                new_path, name = self._create_new_db(name, dbid=dbid, edit_entry=False)

                # Create a new database
                self.__start_cursor(_("Importing data..."))

                dbase = make_database(dbid)
                dbase.load(new_path, user.callback)

                import_function = plugin.get_import_function()
                import_function(dbase, filename, user)

                # finish up
                self.__end_cursor()
                dbase.close()

                return new_path, name
        return None, None

    def is_locked(self, dbpath):
        """
        Returns True if there is a lock file in the dirpath
        """
        if os.path.isfile(os.path.join(dbpath, "lock")):
            return True
        return False

    def needs_recovery(self, dbpath):
        """
        Returns True if the database in dirpath needs recovery
        """
        if os.path.isfile(os.path.join(dbpath, "need_recover")):
            return True
        return False

    def backend_unavailable(self, dbpath):
        """
        Returns True if the database in dirpath has an unavailable backend
        """
        dbid = get_dbid_from_path(dbpath)
        return self.get_backend_name_from_dbid(dbid) == UNAVAILABLE

    def remove_database(self, dbname, user=None):
        """
        Deletes a database folder given a pattenr that matches
        its proper name.
        """
        dbdir = os.path.expanduser(config.get("database.path"))
        match_list = []
        for dpath in os.listdir(dbdir):
            dirpath = os.path.join(dbdir, dpath)
            path_name = os.path.join(dirpath, NAME_FILE)
            if os.path.isfile(path_name):
                with open(path_name, "r", encoding="utf8") as file:
                    name = file.readline().strip()
                if re.match("^" + dbname + "$", name) or dbname == name:
                    match_list.append((name, dirpath))
        if len(match_list) == 0:
            CLIDbManager.ERROR(
                "Family tree not found", "No matching family tree found: '%s'" % dbname
            )
        # now delete them:
        for name, directory in match_list:
            if user is None or user.prompt(
                _("Remove family tree warning"),
                _("Are you sure you want to remove " 'the family tree named\n"%s"?')
                % name,
                _("yes"),
                _("no"),
                default_label=_("no"),
            ):
                try:
                    for top, dirs, files in os.walk(directory):
                        for filename in files:
                            os.unlink(os.path.join(top, filename))
                    os.rmdir(directory)
                except (IOError, OSError) as msg:
                    CLIDbManager.ERROR(_("Could not delete Family Tree"), str(msg))

    def rename_database(self, filepath, new_text):
        """
        Renames the database by writing the new value to the name.txt file
        Returns old_name, new_name if success, None, None if no success
        """
        try:
            with open(filepath, "r", encoding="utf8") as name_file:
                old_text = name_file.read()
            with open(filepath, "w", encoding="utf8") as name_file:
                name_file.write(new_text)
        except (OSError, IOError) as msg:
            CLIDbManager.ERROR(_("Could not rename Family Tree"), str(msg))
            return None, None
        return old_text, new_text

    def break_lock(self, dbpath):
        """
        Breaks the lock on a database
        """
        if os.path.exists(os.path.join(dbpath, "lock")):
            os.unlink(os.path.join(dbpath, "lock"))

    def icon_values(self, dirpath, active, is_open):
        """
        If the directory path is the active path, then return values
        that indicate to use the icon, and which icon to use.
        """
        if os.path.isfile(os.path.join(dirpath, "need_recover")):
            return (True, self.ICON_MAP[self.ICON_RECOVERY])
        elif dirpath == active and is_open:
            return (True, self.ICON_MAP[self.ICON_OPEN])
        elif os.path.isfile(os.path.join(dirpath, "lock")):
            return (True, self.ICON_MAP[self.ICON_LOCK])
        else:
            return (False, self.ICON_MAP[self.ICON_NONE])


def make_dbdir(dbdir):
    """
    Create the default database directory, as defined by dbdir
    """
    try:
        if not os.path.isdir(dbdir):
            os.makedirs(dbdir)
    except (IOError, OSError) as msg:
        LOG.error(
            _(
                "\nERROR: Wrong database path in Edit Menu->Preferences.\n"
                "Open preferences and set correct database path.\n\n"
                "Details: Could not make database directory:\n    %s\n\n"
            ),
            str(msg),
        )
        return False
    return True


def find_next_db_name(name_list):
    """
    Scan the name list, looking for names that do not yet exist.
    Use the DEFAULT_TITLE as the basis for the database name.
    """
    i = 1
    while True:
        title = "%s %d" % (DEFAULT_TITLE, i)
        if title not in name_list:
            return title
        i += 1


def find_next_db_dir():
    """
    Searches the default directory for the first available default
    database name. Base the name off the current time. In all actuality,
    the first should be valid.
    """
    while True:
        base = "%x" % int(time.time())
        dbdir = os.path.expanduser(config.get("database.path"))
        new_path = os.path.join(dbdir, base)
        if not os.path.isdir(new_path):
            break
    return new_path


def time_val(dirpath):
    """
    Return the last modified time of the database. We do this by looking
    at the modification time of the meta db file. If this file does not
    exist, we indicate that database as never modified.
    """
    meta = os.path.join(dirpath, META_NAME)
    if os.path.isfile(meta):
        tval = os.stat(meta)[9]
        # This gives creation date in Windows, but correct date in Linux
        if win():
            # Try to use last modified date instead in Windows
            # and check that it is later than the creation date.
            tval_mod = os.stat(meta)[8]
            if tval_mod > tval:
                tval = tval_mod
        last = time.strftime("%x %X", time.localtime(tval))
    else:
        tval = 0
        last = _("Never")
    return (tval, last)


def find_locker_name(dirpath):
    """
    Opens the lock file if it exists, reads the contexts which is "USERNAME"
    and returns the contents, with correct string before "USERNAME",
    so the message can be printed with correct locale.
    If a file is encountered with errors, we return 'Unknown'
    This data can eg be displayed in the time column of the manager
    """
    try:
        fname = os.path.join(dirpath, "lock")
        with open(fname, "r", encoding="utf8") as ifile:
            username = ifile.read().strip()
            # feature request 2356: avoid genitive form
            last = _("Locked by %s") % username
    except (OSError, IOError, UnicodeDecodeError):
        last = _("Unknown")
    return last



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import *
from gramps.gen.const import URL_MANUAL_PAGE, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"
REPORTS_HELP = URL_MANUAL_PAGE + "_-_Reports#Graphs"

# this is the default in gen/plug/_pluginreg.py: plg.require_active = True

# ------------------------------------------------------------------------
#
# Family Lines Graph
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "familylines_graph"
plg.name = _("Family Lines Graph")
plg.description = _("Produces family line graphs using Graphviz.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "gvfamilylines.py"
plg.ptype = REPORT
plg.authors = ["Stephane Charette"]
plg.authors_email = ["stephanecharette@gmail.com"]
plg.category = CATEGORY_GRAPHVIZ
plg.reportclass = "FamilyLinesReport"
plg.optionclass = "FamilyLinesOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]
plg.require_active = False
plg.help_url = REPORTS_HELP

# ------------------------------------------------------------------------
#
# Hourglass Graph
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "hourglass_graph"
plg.name = _("Hourglass Graph")
plg.description = _("Produces an hourglass graph using Graphviz.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "gvhourglass.py"
plg.ptype = REPORT
plg.authors = ["Brian G. Matherly"]
plg.authors_email = ["brian@gramps-project.org"]
plg.category = CATEGORY_GRAPHVIZ
plg.reportclass = "HourGlassReport"
plg.optionclass = "HourGlassOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]
plg.help_url = REPORTS_HELP

# ------------------------------------------------------------------------
#
# Relationship Graph
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "rel_graph"
plg.name = _("Relationship Graph")
plg.description = _("Produces relationship graphs using Graphviz.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "gvrelgraph.py"
plg.ptype = REPORT
plg.authors = ["Brian G. Matherly"]
plg.authors_email = ["brian@gramps-project.org"]
plg.category = CATEGORY_GRAPHVIZ
plg.reportclass = "RelGraphReport"
plg.optionclass = "RelGraphOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]
plg.help_url = REPORTS_HELP



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    FamilyPage - Family index page and individual Family pages
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from collections import defaultdict, OrderedDict
from decimal import getcontext
import unicodedata
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventType, Family, Name
from gramps.gen.plug.report import Bibliography
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.gen.display.name import displayer as _nd
from gramps.plugins.webreport.common import (
    alphabet_navigation,
    partial_navigation,
    create_indexes_pages,
    html_escape,
    FULLCLEAR,
    AlphabeticIndex,
    get_surname_from_person,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


#################################################
#
#    creates the Family List Page and Family Pages
#
#################################################
class FamilyPages(BasePage):
    """
    This class is responsible for displaying information about the 'Family'
    database objects. It displays this information under the 'Families'
    tab. It is told by the 'add_instances' call which 'Family's to display,
    and remembers the list of Family. A single call to 'display_pages'
    displays both the Family List (Index) page and all the Family
    pages.

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.family_dict = defaultdict(set)
        self.familymappages = None

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Family tab, namely the family
        index and the individual family pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        LOG.debug("obj_dict[Family]")
        for item in self.report.obj_dict[Family].items():
            LOG.debug("    %s", str(item))

        message = _("Creating family pages...")
        index = 1
        progress_title = self.report.pgrs_title(the_lang)
        with self.r_user.progress(
            progress_title, message, len(self.report.obj_dict[Family]) + 1
        ) as step:
            for family_handle in self.report.obj_dict[Family]:
                step()
                index += 1
                self.familypage(self.report, the_lang, the_title, family_handle)
            step()
            self.familylistpage(
                self.report, the_lang, the_title, self.report.obj_dict[Family].keys()
            )

    def __output_family(
        self,
        ldatec,
        family_handle,
        person_handle,
        tbody,
        letter,
        first_person,
        first_family,
    ):
        """
        Generate and output the data for a single family

        @param: ldatec          -- Last change date and time (updated)
        @param: family_handle   -- The family_handle to be output
        @param: person_handle   -- The person_handle to be output
        @param: tbody           -- The current HTML body into which the data is
                                   assembled
        @param: letter          -- The AlphabeticIndex bucket for this event
        @param: first_person    -- Whether this is the first person for this
                                   letter
        @param: first_family    -- Whether this is the first family of this
                                   person

        @returns: Returns a tuple of updated (ldatec, first_person,
                                              first_family)
        @rtype: tuple
        """
        family = self.r_db.get_family_from_handle(family_handle)
        if family.get_change_time() > ldatec:
            ldatec = family.get_change_time()

        trow = Html("tr")
        tbody += trow
        tcell = Html("td", class_="ColumnRowLabel")
        trow += tcell
        if first_person:
            first_person = False
            first_family = False
            # Update the ColumnRowLabel cell
            trow.attr = 'class="BeginLetter BeginFamily"'
            ttle = self._("Families beginning with " "letter ")
            tcell += Html("a", letter, name=letter, title=ttle + letter)
            #  and create the populated ColumnPartner for the person
            tcell = Html("td", class_="ColumnPartner")
            tcell += self.new_person_link(person_handle, uplink=self.uplink)
            trow += tcell
        elif first_family:
            first_family = False
            # Update the ColumnRowLabel cell
            trow.attr = 'class ="BeginFamily"'
            #  and create the populated ColumnPartner for the person
            tcell = Html("td", class_="ColumnPartner")
            tcell += self.new_person_link(person_handle, uplink=self.uplink)
            trow += tcell
        else:
            # Create the blank ColumnPartner row for the person
            tcell = Html("td", class_="ColumnPartner")
            tcell += "&nbsp;"
            trow += tcell

        tcell = Html("td", class_="ColumnPartner")
        trow += tcell
        tcell += self.family_link(
            family.get_handle(),
            self.report.get_family_name(family),
            family.get_gramps_id(),
            self.uplink,
        )
        # family events; such as marriage and divorce
        # events
        fam_evt_ref_list = family.get_event_ref_list()
        tcell1 = Html("td", class_="ColumnDate", inline=True)
        tcell2 = Html("td", class_="ColumnDate", inline=True)
        trow += tcell1, tcell2
        if fam_evt_ref_list:
            fam_evt_srt_ref_list = sorted(fam_evt_ref_list, key=self.sort_on_grampsid)
            for evt_ref in fam_evt_srt_ref_list:
                evt = self.r_db.get_event_from_handle(evt_ref.ref)
                if evt:
                    evt_type = evt.get_type()
                    if evt_type in [EventType.MARRIAGE, EventType.DIVORCE]:
                        cell = self.rlocale.get_date(evt.get_date_object())
                        if evt_type == EventType.MARRIAGE:
                            tcell1 += cell
                        else:
                            tcell1 += "&nbsp;"
                        if evt_type == EventType.DIVORCE:
                            tcell2 += cell
                        else:
                            tcell2 += "&nbsp;"

        else:
            tcell1 += "&nbsp;"
            tcell2 += "&nbsp;"
        first_family = False
        return (ldatec, first_person, first_family)

    def part_familylistpage(
        self,
        report,
        index_list,
        name,
        letter,
        family_handle_list,
        part=None,
        subp=None,
        partial_list=None,
    ):
        """
        Create a partial family index
        """
        nav_name = part_name = name
        if part != 0 and subp != 0:
            part_name += str(part)
            nav_name = part_name
        if subp and subp != 0:
            part_name += subp
        output_file, sio = self.report.create_file(part_name)
        result = self.write_header(self._("Families"))
        familieslistpage, dummy_head, dummy_body, outerwrapper = result
        ldatec = 0

        # begin Family Division
        with Html("div", class_="content", id="Relationships") as relationlist:
            outerwrapper += relationlist

            # Families list page message
            msg = self._(
                "This page contains an index of all the "
                "families/ relationships in the "
                "database, sorted by their family name/ surname. "
                "Clicking on a person&#8217;s "
                "name will take you to their "
                "family/ relationship&#8217;s page."
            )
            relationlist += Html("p", msg, id="description")

            # Output the navigation
            # add alphabet navigation
            alpha_nav = alphabet_navigation(
                index_list,
                self.rlocale,
                current=part_name,
                rtl=self.dir,
                new_page="families",
                ext=self.ext,
            )
            if alpha_nav:
                relationlist += alpha_nav

            if part is not None:
                # We need to create a new navigation tab for partial page index.
                partial_nav = partial_navigation(
                    partial_list,
                    self.rlocale,
                    current=part_name,
                    rtl=self.dir,
                    ext=self.ext,
                    new_page=nav_name,
                )
                if partial_nav is not None:
                    relationlist += Html(
                        "div", style="clear: both;"
                    )  # This to align the next div to the left.
                    relationlist += partial_nav

            # begin families table and table head
            with Html("table", class_="infolist relationships " + self.dir) as table:
                relationlist += table

                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                # set up page columns
                trow.extend(
                    Html("th", trans, class_=colclass, inline=True)
                    for trans, colclass in [
                        (self._("Letter"), "ColumnRowLabel"),
                        (self._("Person"), "ColumnPartner"),
                        (self._("Family"), "ColumnPartner"),
                        (self._("Marriage"), "ColumnDate"),
                        (self._("Divorce"), "ColumnDate"),
                    ]
                )

                tbody = Html("tbody")
                table += tbody

                first_person = True
                # family_handle_list = family_handle_list[0][1]
                for person_handle, surname in family_handle_list:
                    person = self.r_db.get_person_from_handle(person_handle)
                    if person:
                        family_list = person.get_family_handle_list()
                        first_family = True
                        for family_handle in family_list:
                            (
                                ldatec,
                                first_person,
                                first_family,
                            ) = self.__output_family(
                                ldatec,
                                family_handle,
                                person_handle,
                                tbody,
                                letter,
                                first_person,
                                first_family,
                            )

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(familieslistpage, output_file, sio, ldatec)

    def familylistpage(self, report, the_lang, the_title, fam_list):
        """
        Create a family index

        @param: report    -- The instance of the main report class for
                             this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        @param: fam_list  -- The handle for the place to add
        """
        BasePage.__init__(self, report, the_lang, the_title)
        ldatec = 0

        pers_fam_dict = defaultdict(list)
        for family_handle in fam_list:
            family = self.r_db.get_family_from_handle(family_handle)
            if family:
                if family.get_change_time() > ldatec:
                    ldatec = family.get_change_time()
                husband_handle = family.get_father_handle()
                spouse_handle = family.get_mother_handle()
                if husband_handle:
                    pers_fam_dict[husband_handle].append(family)
                if spouse_handle:
                    pers_fam_dict[spouse_handle].append(family)

        # Assemble all the people, we no longer care about their families
        index = AlphabeticIndex(self.rlocale)
        for person_handle, dummy_family in pers_fam_dict.items():
            person = self.r_db.get_person_from_handle(person_handle)
            surname = get_surname_from_person(self.r_db, person)
            index.addRecord(surname, person_handle)

        # Extract the buckets from the index
        index_list = []
        index.resetBucketIterator()
        # for each bucket, output the people and their families in that
        # bucket
        surname_ppl_handle_dict = OrderedDict()
        while index.nextBucket():
            if index.bucketRecordCount != 0:
                letter = index.bucketLabel
                # Assemble a dict of all the people in this bucket.
                bletter = (
                    unicodedata.normalize("NFKD", letter)[0]
                    if len(letter) > 0
                    else letter
                )
                while index.nextRecord():
                    # The records are returned sorted by recordName,
                    # which is surname. we need to retain that order but
                    # in addition sort by the rest of the name
                    person_surname = index.recordName
                    person_handle = index.recordData
                    if person_surname in surname_ppl_handle_dict.keys():
                        surname_ppl_handle_dict[person_surname].append(person_handle)
                    else:
                        surname_ppl_handle_dict[person_surname] = [person_handle]
        # sort by surname
        surname_handle_list = surname_ppl_handle_dict
        extended_handle_list = defaultdict(list)
        name_format = self.report.options["name_format"]
        max_letter_rows = defaultdict(int)
        index_list = []
        for bletter, hdlel in surname_handle_list.items():
            bletter = (
                unicodedata.normalize("NFKD", bletter)[0]
                if len(bletter) > 0
                else bletter
            )
            bletter = bletter[0] if len(bletter) > 0 else bletter
            bletter = bletter.upper() if bletter.isalpha() else "…"
            if bletter not in index_list:
                index_list.append(bletter)
            max_letter_rows[bletter] = len(hdlel)
            for hdle in hdlel:
                person = self.r_db.get_person_from_handle(hdle)
                primary_name = person.get_primary_name()
                nname = Name(primary_name)
                nname.set_display_as(name_format)
                fname = _nd.display_name(nname)
                extended_handle_list[bletter].append((hdle, fname))
        extended_handle_list = list(extended_handle_list.items())
        create_indexes_pages(
            report,
            "families",
            index_list,
            self.part_familylistpage,
            extended_handle_list,
            locale=self.rlocale,
        )

    def familypage(self, report, the_lang, the_title, family_handle):
        """
        Create a family page

        @param: report        -- The instance of the main report class for
                                 this report
        @param: the_lang      -- The lang to process
        @param: the_title     -- The title page related to the language
        @param: family_handle -- The handle for the family to add
        """
        family = report.database.get_family_from_handle(family_handle)
        if not family:
            return
        BasePage.__init__(self, report, the_lang, the_title, family.get_gramps_id())
        ldatec = family.get_change_time()

        self.bibli = Bibliography()
        self.uplink = True
        family_name = self.report.get_family_name(family)
        self.page_title = family_name

        self.familymappages = report.options["familymappages"]

        output_file, sio = self.report.create_file(family.get_handle(), "fam")
        result = self.write_header(family_name)
        familydetailpage, head, dummy_body, outerwrapper = result

        # begin FamilyDetaill division
        with Html(
            "div", class_="content", id="RelationshipDetail"
        ) as relationshipdetail:
            outerwrapper += relationshipdetail

            # family media list for initial thumbnail
            if self.create_media:
                media_list = family.get_media_list()
                if media_list:
                    if self.the_lang and not self.usecms:
                        fname = "/".join(["..", "css", "lightbox.css"])
                        jsname = "/".join(["..", "css", "lightbox.js"])
                    else:
                        fname = "/".join(["css", "lightbox.css"])
                        jsname = "/".join(["css", "lightbox.js"])
                    url = self.report.build_url_fname(fname, None, self.uplink)
                    head += Html(
                        "link",
                        href=url,
                        type="text/css",
                        media="screen",
                        rel="stylesheet",
                    )
                    url = self.report.build_url_fname(jsname, None, self.uplink)
                    head += Html("script", src=url, type="text/javascript", inline=True)
                # If Event pages are not being created, then we need to display
                # the family event media here
                if not self.inc_events:
                    for evt_ref in family.get_event_ref_list():
                        event = self.r_db.get_event_from_handle(evt_ref.ref)
                        media_list += event.get_media_list()

            relationshipdetail += Html("h2", self.page_title, inline=True) + (
                Html("sup")
                + (Html("small") + self.get_citation_links(family.get_citation_list()))
            )
            # Tags
            tags = self.show_tags(family)
            if tags and self.report.inc_tags:
                trow = Html("table", class_="tags") + (
                    Html("tr")
                    + (
                        Html(
                            "td",
                            self._("Tags") + self._(":"),
                            class_="ColumnAttribute",
                            inline=True,
                        ),
                        Html("td", tags, class_="ColumnValue", inline=True),
                    )
                )
                relationshipdetail += trow

            # display relationships
            families = self.display_family_relationships(family, None)
            if families is not None:
                relationshipdetail += families

            # display additional images as gallery
            if self.create_media:
                addgallery = self.disp_add_img_as_gallery(media_list, family)
                if addgallery:
                    relationshipdetail += addgallery

            # Narrative subsection
            notelist = family.get_note_list()
            if notelist:
                relationshipdetail += self.display_note_list(notelist, Family)

            # display family LDS ordinance...
            family_lds_ordinance_list = family.get_lds_ord_list()
            if family_lds_ordinance_list:
                relationshipdetail += self.display_lds_ordinance(family)

            # get attribute list
            attrlist = family.get_attribute_list()
            if attrlist:
                attrsection, attrtable = self.display_attribute_header()
                self.display_attr_list(attrlist, attrtable)
                relationshipdetail += attrsection

            # for use in family map pages...
            if self.report.options["familymappages"]:
                name_format = self.report.options["name_format"]
                father = mother = None
                with self.create_toggle("map") as h4_head:
                    relationshipdetail += h4_head
                    h4_head += self._("Family Map")
                    disp = "none" if self.report.options["toggle"] else "block"
                    with Html(
                        "div", style="display:%s" % disp, id="toggle_map"
                    ) as toggle:
                        relationshipdetail += toggle
                        mapdetail = Html("br")
                        fhdle = family.get_father_handle()
                        for handle, dummy_url in self.report.fam_link.items():
                            if fhdle == handle:
                                father = self.r_db.get_person_from_handle(fhdle)
                                break
                        if father:
                            primary_name = father.get_primary_name()
                            name = Name(primary_name)
                            name.set_display_as(name_format)
                            fname = html_escape(_nd.display_name(name))
                            mapdetail += self.family_map_link_for_parent(fhdle, fname)
                        mapdetail += Html("br")
                        mhdle = family.get_mother_handle()
                        for handle, dummy_url in self.report.fam_link.items():
                            if mhdle == handle:
                                mother = self.r_db.get_person_from_handle(mhdle)
                                break
                        if mother:
                            primary_name = mother.get_primary_name()
                            name = Name(primary_name)
                            name.set_display_as(name_format)
                            mname = html_escape(_nd.display_name(name))
                            mapdetail += self.family_map_link_for_parent(mhdle, mname)
                        toggle += mapdetail

            # source references
            srcrefs = self.display_ind_sources(family)
            if srcrefs:
                relationshipdetail += srcrefs

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(familydetailpage, output_file, sio, ldatec)

    def family_map_link_for_parent(self, handle, name):
        """
        Creates a link to the family map for the father or the mother

        @param: handle -- The person handle
        @param: name   -- The name for this person to display
        """
        url = self.report.fam_link[handle]
        title = self._("Family Map for %s") % name
        return Html("a", title, href=url, title=title, class_="family_map", inline=True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# test/GrampsLogger/GtkHandler_Test.py

import os
import unittest
import logging
import sys

from gi.repository import Gtk

log = logging.getLogger("Gramps.Tests.GrampsLogger")
import gramps.gen.const as const

rootDir = os.path.join(os.path.dirname(__file__), "../../gramps")
sys.path.append(os.path.join(rootDir, "test"))
sys.path.append(rootDir)

from gramps.gui.logger import RotateHandler, GtkHandler


class GtkHandlerTest(unittest.TestCase):
    """Test the GtkHandler."""

    def test_window(self):
        """Test that the window appears."""

        rh = RotateHandler(capacity=20)
        rh.setLevel(logging.DEBUG)

        gtkh = GtkHandler(rotate_handler=rh)
        gtkh.setLevel(logging.ERROR)

        l = logging.getLogger("GtkHandlerTest")
        l.setLevel(logging.DEBUG)

        l.addHandler(rh)
        l.addHandler(gtkh)

        l.info("An info message")
        l.warning("A warn message")
        l.debug("A debug message")
        log_message = "Debug message"


# Comment this out because there is noone to close the dialogue
#         try:
#             wibble
#         except:
#             l.error(log_message,exc_info=True)
#
#         while Gtk.events_pending():
#             Gtk.main_iteration()


def testSuite():
    suite = unittest.TestLoader().loadTestsFromTestCase(GtkHandlerTest)
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner().run(testSuite())



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Vassilii Khachaturov <vassilii@tarunz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for user.py"""

import unittest
from unittest.mock import Mock, patch
from .. import user


class TestUser:
    TITLE = "Testing prompt"
    MSG = "Choices are hard. Nevertheless, please choose!"
    ACCEPT = "To be"
    REJECT = "Not to be"


class TestUser_prompt(unittest.TestCase):
    def setUp(self):
        self.user = user.User()

    def test_prompt_runs_QuestionDialog2(self):
        with patch("gramps.gui.user.QuestionDialog2") as MockQD:
            self.user.prompt(
                TestUser.TITLE,
                TestUser.MSG,
                TestUser.ACCEPT,
                TestUser.REJECT,
                parent=None,
            )
        MockQD.assert_called_once_with(
            TestUser.TITLE, TestUser.MSG, TestUser.ACCEPT, TestUser.REJECT, parent=None
        )
        MockQD.return_value.run.assert_called_once_with()
        # TODO test that run's return is the one returned by prompt()...


class TestUser_init_accepts_uistate(unittest.TestCase):
    def test(self):
        user.User(uistate=None)


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Address List display tab.
"""

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gi.repository import GObject, GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Address
from gramps.gen.errors import WindowActiveError
from ...ddtargets import DdTargets
from .addressmodel import AddressModel
from .embeddedlist import EmbeddedList, TEXT_COL, MARKUP_COL, ICON_COL


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class AddrEmbedList(EmbeddedList):
    """
    Address List display tab for edit dialogs.

    Derives from the EmbeddedList class.
    """

    _HANDLE_COL = 10
    _DND_TYPE = DdTargets.ADDRESS

    _MSG = {
        "add": _("Create and add a new address"),
        "del": _("Remove the existing address"),
        "edit": _("Edit the selected address"),
        "up": _("Move the selected address upwards"),
        "down": _("Move the selected address downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("Date"), 0, 150, MARKUP_COL, -1, None),
        (_("Street"), 1, 225, TEXT_COL, -1, None),
        (_("Locality"), 2, 100, TEXT_COL, -1, None),
        (_("City"), 3, 100, TEXT_COL, -1, None),
        (_("State/County"), 4, 100, TEXT_COL, -1, None),
        (_("Country"), 5, 75, TEXT_COL, -1, None),
        (_("Postal"), 6, 75, TEXT_COL, -1, None),
        (_("Phone"), 7, 150, TEXT_COL, -1, None),
        (_("Source"), 8, 30, ICON_COL, -1, "gramps-source"),
        (_("Private"), 9, 30, ICON_COL, -1, "gramps-lock"),
    ]

    def __init__(self, dbstate, uistate, track, data, config_key):
        self.data = data
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Addresses"),
            AddressModel,
            config_key,
            move_buttons=True,
        )

    def get_icon_name(self):
        """
        Return the icon name associated with the display tab
        """
        return "gramps-address"

    def get_data(self):
        """
        Return the data associated with display tab
        """
        return self.data

    def column_order(self):
        """
        Return the column order of the columns in the display tab.
        """
        return (
            (1, 8),
            (1, 9),
            (1, 0),
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (1, 5),
            (1, 6),
            (1, 7),
        )

    def add_button_clicked(self, obj):
        """
        Called when the Add button is clicked. Creates a new Address instance
        and calls the EditAddress editor with the new address. If the window
        already exists (WindowActiveError), we ignore it. This prevents
        the dialog from coming up twice on the same object.
        """
        addr = Address()
        try:
            from .. import EditAddress

            EditAddress(self.dbstate, self.uistate, self.track, addr, self.add_callback)
        except WindowActiveError:
            return

    def add_callback(self, name):
        """
        Called to update the screen when a new address is added
        """
        data = self.get_data()
        data.append(name)
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, len(data) - 1)

    def edit_button_clicked(self, obj):
        """
        Called with the Edit button is clicked. Gets the selected Address instance
        and calls the EditAddress editor with the address. If the window
        already exists (WindowActiveError), we ignore it. This prevents
        the dialog from coming up twice on the same object.
        """
        addr = self.get_selected()
        if addr:
            try:
                from .. import EditAddress

                EditAddress(
                    self.dbstate, self.uistate, self.track, addr, self.edit_callback
                )
            except WindowActiveError:
                return

    def edit_callback(self, name):
        """
        Called to update the screen when the address changes
        """
        self.rebuild()
